#!/usr/bin/python3

import argparse
import datetime
import time
import json
import os
import re
import sys
import xml.etree.ElementTree as ET

import ssg.build_yaml
import ssg.constants
import ssg.rules
import ssg.yaml
import ssg.environment

SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
SSG_BUILD_ROOT = os.path.join(SSG_ROOT, "build")
RULES_JSON = os.path.join(SSG_BUILD_ROOT, "rule_dirs.json")
BUILD_CONFIG = os.path.join(SSG_BUILD_ROOT, "build_config.yml")
NS = {'scap': ssg.constants.datastream_namespace,
      'xccdf-1.2': ssg.constants.XCCDF12_NS}
PROFILE = 'stig'


def get_profile(product, profile_name):
    ds_root = ET.parse(os.path.join(SSG_ROOT, 'build', 'ssg-{product}-ds.xml'
                                    .format(product=product))).getroot()
    profiles = ds_root.findall(
        './/{{{scap}}}component/{{{xccdf}}}Benchmark/{{{xccdf}}}Profile'.format(
            scap=NS["scap"], xccdf=NS["xccdf-1.2"])
    )

    profile_name_fqdn = "xccdf_org.ssgproject.content_profile_{profile_name}".format(
        profile_name=profile_name)

    for profile in profiles:
        if profile.attrib['id'] == profile_name_fqdn:
            return profile
    raise ValueError("Profile %s was not found." % profile_name_fqdn)


get_profile.__annotations__ = {'product': str, 'profile_name': str, 'return': ET.Element}


def filter_out_implemented_rules(known_rules, ns, root):
    needed_rules = known_rules.copy()
    groups = root.findall('.//scap:component/xccdf-1.2:Benchmark/xccdf-1.2:Group', ns)
    for group in groups:
        for stig in group.findall('xccdf-1.2:Rule', ns):
            stig_id = stig.find('xccdf-1.2:version', ns).text
            check = stig.find('xccdf-1.2:check', ns)
            if stig_id in known_rules.keys() and len(check) > 0:
                del needed_rules[stig_id]
    return needed_rules


filter_out_implemented_rules.__annotations__ = {'known_rules': dict, 'ns': dict,
                                                'root': ET.Element, 'return': dict}


def handle_rule_yaml(product, rule_id, rule_dir, guide_dir, env_yaml):
    rule_obj = {'id': rule_id, 'dir': rule_dir, 'guide': guide_dir}
    rule_file = ssg.rules.get_rule_dir_yaml(rule_dir)

    rule_yaml = ssg.build_yaml.Rule.from_yaml(rule_file, env_yaml=env_yaml)
    rule_yaml.normalize(product)
    rule_obj['references'] = rule_yaml.references
    return rule_obj


handle_rule_yaml.__annotations__ = {'product': str, 'rule_id': str, 'rule_dir': str,
                                    'guide_dir': str, 'env_yaml': dict, 'return': dict}


def get_platform_rules(product, json_path, resolved_rules_dir, build_root):
    platform_rules = list()
    if resolved_rules_dir:
        rules_path = os.path.join(build_root, product, 'rules')
        for file in os.listdir(rules_path):
            if not file.endswith('.yml'):
                continue
            rule_id = file.replace('.yml', '')
            rule = dict()
            rule_yaml = ssg.yaml.open_raw(os.path.join(rules_path, file))
            rule['id'] = rule_id
            rule['references'] = dict()
            rule['references'] = rule_yaml['references']
            platform_rules.append(rule)
    else:
        rules_json_file = open(json_path, 'r')
        rules_json = json.load(rules_json_file)
        for rule in rules_json.values():
            if product in rule['products']:
                platform_rules.append(rule)
        if not rules_json_file.closed:
            rules_json_file.close()
    return platform_rules


get_platform_rules.__annotations__ = {'product': str, 'json_path': str, 'resolved_rules_dir': bool,
                                      'build_root': str, 'return': list}


def _open_rule_obj(resolved_rules_dir, rule, product, env_yaml):
    if resolved_rules_dir:
        return rule
    try:
        rule_obj = handle_rule_yaml(
            product, rule['id'], rule['dir'], rule['guide'], env_yaml)
    except ssg.yaml.DocumentationNotComplete:
        msg = 'Rule %s throw DocumentationNotComplete' % rule['id']
        sys.stderr.write(msg)
        # Happens on non-debug build when a rule is "documentation-incomplete"
        return None
    return rule_obj


def get_implemented_stigs(product, root_path, build_config_yaml_path,
                          reference_str, json_path, resolved_rules_dir,
                          build_root):
    platform_rules = get_platform_rules(product, json_path, resolved_rules_dir, build_root)

    product_dir = os.path.join(root_path, "products", product)
    product_yaml_path = os.path.join(product_dir, "product.yml")
    env_yaml = ssg.environment.open_environment(
        build_config_yaml_path, product_yaml_path, os.path.join(root_path, "product_properties"))

    known_rules = dict()
    for rule in platform_rules:
        rule_obj = _open_rule_obj(resolved_rules_dir, rule, product, env_yaml)
        if not rule_obj:
            continue
        if reference_str not in rule_obj['references'].keys():
            continue
        refs = rule_obj['references'][reference_str]
        for ref in refs:
            if ref in known_rules:
                known_rules[ref].append(rule['id'])
            else:
                known_rules[ref] = [rule['id']]
    return known_rules


get_implemented_stigs.__annotations__ = {'product': str, 'root_path': str,
                                         'build_config_yaml_path': str, 'reference_str': str,
                                         'json_path': str, 'resolved_rules_dir': bool,
                                         'build_root': str, 'return': dict}


def setup_tailoring_profile(profile_id, profile_root):
    tailoring_profile = ET.Element('xccdf-1.2:Profile')
    if profile_id:
        tailoring_profile.set('id', '{oscap}{profile_id}'
                              .format(oscap=ssg.constants.OSCAP_PROFILE, profile_id=profile_id))
    else:
        tailoring_profile.set('id', '{id}_delta_tailoring'.format(id=profile_root.attrib["id"]))
    tailoring_profile.append(profile_root.find('xccdf-1.2:title', NS))
    tailoring_profile.append(profile_root.find('xccdf-1.2:description', NS))
    tailoring_profile.set('extends', profile_root.get('id'))
    return tailoring_profile


setup_tailoring_profile.__annotations__ = {'profile_id': str, 'profile_root': ET.Element}


def _get_datetime():
    return datetime.datetime.fromtimestamp(
        int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))).isoformat()


def create_tailoring(args):
    benchmark_root = ET.parse(args.manual).getroot()
    known_rules = get_implemented_stigs(args.product, args.root, args.build_config_yaml,
                                        args.reference, args.json, args.resolved_rules_dir,
                                        args.build_root)
    needed_rules = filter_out_implemented_rules(known_rules, NS, benchmark_root)
    needed_rule_names_set = set(rulename for ruleset in needed_rules.values() for rulename in ruleset)
    profile_root = get_profile(args.product, args.profile)
    selections = profile_root.findall('xccdf-1.2:select', NS)
    tailoring_profile = setup_tailoring_profile(args.profile_id, profile_root)
    for selection in selections:
        if selection.attrib['idref'].startswith(ssg.constants.OSCAP_RULE):
            cac_rule_id = selection.attrib['idref'].replace(ssg.constants.OSCAP_RULE, '')
            desired_value = str(cac_rule_id in needed_rule_names_set).lower()
            if selection.get('selected') != desired_value:
                selection.set('selected', desired_value)
                tailoring_profile.append(selection)
                if not args.quiet:
                    print('Set rule "{cac_rule_id}" selection state to {desired_value}'
                          .format(cac_rule_id=cac_rule_id, desired_value=desired_value))

    tailoring_root = ET.Element('xccdf-1.2:Tailoring')
    version = ET.SubElement(tailoring_root, 'xccdf-1.2:version',
                            attrib={'time': _get_datetime()})
    version.text = '1'
    tailoring_root.set('id', args.tailoring_id)
    tailoring_root.append(tailoring_profile)
    return tailoring_root


create_tailoring.__annotations__ = {'args': argparse.Namespace, 'return': ET.Element}


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--root", type=str, action="store", default=SSG_ROOT,
                        help="Path to SSG root directory (defaults to {ssg_root})"
                        .format(ssg_root=SSG_ROOT))
    parser.add_argument("-p", "--product", type=str, action="store", required=True,
                        help="What product to produce the tailoring file for")
    parser.add_argument("-m", "--manual", type=str, action="store", required=True,
                        help="Path to XML XCCDF manual file to use as the source")
    parser.add_argument("-j", "--json", type=str, action="store", default=RULES_JSON,
                        help="Path to the rules_dir.json (defaults to build/stig_control.json)")
    parser.add_argument("-c", "--build-config-yaml", default=BUILD_CONFIG,
                        help="YAML file with information about the build configuration. ")
    parser.add_argument("-b", "--profile", type=str, default="stig",
                        help="What profile to use. Defaults to stig")
    parser.add_argument("-ref", "--reference", type=str, default="stigid",
                        help="Reference system to check for, defaults to stigid")
    parser.add_argument("-o", "--output", type=str,
                        help="Defaults build/PRODUCT_PROFILE_tailoring.xml")
    parser.add_argument("--profile-id", type=str,
                        help="Id of the created profile. Defaults to PROFILE_delta")
    parser.add_argument("--tailoring-id", type=str,
                        default='xccdf_content-disa-delta_tailoring_default',
                        help="Id of the created tailoring file. "
                             "Defaults to xccdf_content-disa-delta_tailoring_default")
    parser.add_argument("-q", "--quiet", action='store_true',
                        help="The script will not produce any output.")
    parser.add_argument("--resolved-rules-dir", action='store_true',
                        help="The script will not use rules_dir.json, "
                             "but instead uses the data from the build.")
    parser.add_argument('-B', '--build-root', type=str, default=SSG_BUILD_ROOT,
                        help="The root of the CMake working directory, "
                             "defaults to {ssg_build_root}".format(ssg_build_root=SSG_BUILD_ROOT))
    parser.add_argument("-d", "--dry-run", action="store_true",
                        help="If set the script will not output.")
    return parser.parse_args()


parse_args.__annotations__ = {'return': argparse.Namespace}


def main():
    args = parse_args()
    ET.register_namespace('xccdf-1.2', ssg.constants.XCCDF12_NS)
    tailoring_root = create_tailoring(args)
    tree = ET.ElementTree(tailoring_root)
    manual_version = re.search(r'(v[0-9]+r[0-9]+)', args.manual)
    if manual_version is None:
        sys.stderr.write("Unable to find version from file name.\n")
        sys.stderr.write("The string v[NUM]r[NUM] must be in the filename.\n")
        exit(1)

    if args.dry_run:
        return 0

    if args.output:
        out = os.path.join(args.output)
    else:
        out = os.path.join(SSG_ROOT, 'build',
                           '{product}_{profile}_{manual_version}_delta_tailoring.xml'
                           .format(product=args.product, profile=args.profile,
                                   manual_version=manual_version.group(0)))
    tree.write(out)
    if not args.quiet:
        print("Wrote tailoring file to {out}.".format(out=out))


if __name__ == '__main__':
    main()



"""
Common functions for processing YAML in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import codecs
import re
import sys
import yaml

from collections import OrderedDict

from .jinja import (
    load_macros,
    load_macros_from_content_dir,
    process_file,
)

try:
    from yaml import CSafeLoader as yaml_SafeLoader
except ImportError:
    from yaml import SafeLoader as yaml_SafeLoader

try:
    from yaml import CLoader as yaml_Loader
except ImportError:
    from yaml import Loader as yaml_Loader

try:
    from yaml import CDumper as yaml_Dumper
except ImportError:
    from yaml import Dumper as yaml_Dumper

def _bool_constructor(self, node):
    """
    Construct a boolean value from a YAML node.

    Args:
        self: The instance of the class calling this method.
        node: The YAML node to be converted into a boolean value.

    Returns:
        The scalar value of the node, typically a boolean.
    """
    return self.construct_scalar(node)


def _unicode_constructor(self, node):
    """
    Constructs a Unicode string from a YAML node.

    This method takes a YAML node, extracts its scalar value, and converts it to a Unicode string.

    Args:
        node (yaml.Node): The YAML node to be converted.

    Returns:
        str: The Unicode string representation of the node's scalar value.
    """
    string_like = self.construct_scalar(node)
    return str(string_like)


# Don't follow python bool case
yaml_SafeLoader.add_constructor(u'tag:yaml.org,2002:bool', _bool_constructor)
# Python2-relevant - become able to resolve "unicode strings"
yaml_SafeLoader.add_constructor(u'tag:yaml.org,2002:python/unicode', _unicode_constructor)


class DocumentationNotComplete(Exception):
    pass


def _save_rename(result, stem, prefix):
    """
    Renames a key (represented by the stem argument) in the result dictionary by prefixing it with a given prefix.

    Args:
        result (dict): The dictionary where the renamed key-value pair will be stored.
        stem (str): The original key that will be renamed.
        prefix (str): The prefix to be added to the original key.

    Returns:
        None
    """
    result["{0}_{1}".format(prefix, stem)] = stem


def _get_yaml_contents_without_documentation_complete(parsed_yaml, substitutions_dict):
    """
    Processes the given parsed YAML content by handling the 'documentation_complete' key.

    If the YAML content is a dictionary, it checks the value of the 'documentation_complete' key.
    If 'documentation_complete' is "false" and the build type is not "Debug", it raises a
    DocumentationNotComplete exception. The 'documentation_complete' key is then removed from the
    dictionary.

    If the YAML content is empty or a list, it is returned as is.

    Args:
        parsed_yaml (dict or list): The parsed YAML content.
        substitutions_dict (dict): A dictionary containing substitution values, including the
                                   'cmake_build_type'.

    Returns:
        dict or list: The processed YAML content with the 'documentation_complete' key removed if
                      applicable.

    Raises:
        DocumentationNotComplete: If 'documentation_complete' is "false" and the build type is not
                                  "Debug".
    """
    if isinstance(parsed_yaml, dict):
        documentation_incomplete_content_and_not_debug_build = (
            parsed_yaml.pop("documentation_complete", "true") == "false"
            and substitutions_dict.get("cmake_build_type") != "Debug")
        if documentation_incomplete_content_and_not_debug_build:
            raise DocumentationNotComplete("documentation not complete and not a debug build")
    return parsed_yaml


def _open_yaml(stream, original_file=None, substitutions_dict={}):
    """
    Open given file-like object and parse it as YAML.

    Args:
        stream (file-like object): The file-like object containing YAML content.
        original_file (str, optional): The path to the original file for better error handling.
                                       Defaults to None.
        substitutions_dict (dict, optional): A dictionary of substitutions to apply to the YAML
                                             content. Defaults to {}.

    Returns:
        dict: The parsed YAML content with substitutions applied.

    Raises:
        DocumentationNotComplete: If the YAML content contains the key "documentation_complete"
                                  set to "false".
        Exception: For any other exceptions, including tab indentation errors in the file.
    """
    try:
        yaml_contents = yaml.load(stream, Loader=yaml_SafeLoader)

        return _get_yaml_contents_without_documentation_complete(yaml_contents, substitutions_dict)
    except DocumentationNotComplete as e:
        raise e
    except Exception as e:
        count = 0
        _file = original_file
        if not _file:
            _file = stream
        with open(_file, "r") as e_file:
            lines = e_file.readlines()
            for line in lines:
                count = count + 1
                if re.match(r"^\s*\t+\s*", line):
                    print("Exception while handling file: %s" % _file, file=sys.stderr)
                    print("TabIndentationError: Line %s contains tabs instead of spaces:" % (count), file=sys.stderr)
                    print("%s\n\n" % repr(line.strip("\n")), file=sys.stderr)
                    sys.exit(1)

        print("Exception while handling file: %s" % _file, file=sys.stderr)
        raise e


def open_and_expand(yaml_file, substitutions_dict=None):
    """
    Process the given YAML file as a template, using the provided substitutions dictionary to
    perform variable expansion. After expanding the template, process the result as YAML content.

    Args:
        yaml_file (str): The path to the YAML file to be processed.
        substitutions_dict (dict, optional): A dictionary containing key-value pairs for template
            substitution. Defaults to an empty dictionary if not provided.

    Returns:
        dict: The processed YAML content as a dictionary.

    Raises:
        yaml.scanner.ScannerError: If there is an error in scanning the YAML content, typically
            due to incorrect indentation after template expansion.

    See also:
        _open_yaml: Function to open and parse the YAML content.
    """
    if substitutions_dict is None:
        substitutions_dict = dict()

    expanded_template = process_file(yaml_file, substitutions_dict)
    try:
        yaml_contents = _open_yaml(expanded_template, yaml_file, substitutions_dict)
    except yaml.scanner.ScannerError as e:
        print("A Jinja template expansion can mess up the indentation.")
        print("Please, check if the contents below are correctly expanded:")
        print("Source yaml: {}".format(yaml_file))
        print("Expanded yaml:\n{}".format(expanded_template))
        sys.exit(1)

    return yaml_contents


def open_and_macro_expand(yaml_file, substitutions_dict=None):
    """
    Opens a YAML file and expands macros within it using the provided substitutions dictionary.

    This function loads definitions of macros and uses them to expand the template within the YAML
    file. It is similar to open_and_expand, but load definitions of macros

    Args:
        yaml_file (str): The path to the YAML file to be opened and expanded.
        substitutions_dict (dict, optional): A dictionary containing macro definitions and their
                                             corresponding values. If not provided, an empty
                                             dictionary is used.

    Returns:
        dict: The expanded content of the YAML file with macros substituted.
    """
    substitutions_dict = load_macros(substitutions_dict)
    return open_and_expand(yaml_file, substitutions_dict)


def open_and_macro_expand_from_dir(yaml_file, content_dir, substitutions_dict=None):
    """
    Opens a YAML file and expands macros from a specified directory. It is similar to
    open_and_macro_expand but loads macro definitions from a specified directory instead of the
    default directory defined in constants. This is useful in cases where the SSG library is
    consumed by an external project.

    Args:
        yaml_file (str): The path to the YAML file to be opened and expanded.
        content_dir (str): The content dir directory to be used for expansion.
        substitutions_dict (dict, optional): A dictionary of substitutions to be used for macro
                                             expansion. If None, a new dictionary will be created
                                             from the content_dir.

    Returns:
        dict: The expanded content of the YAML file.
    """
    substitutions_dict = load_macros_from_content_dir(content_dir, substitutions_dict)
    return open_and_expand(yaml_file, substitutions_dict)


def open_raw(yaml_file):
    """
    Open the given YAML file and parse its contents without performing any template processing.

    Args:
        yaml_file (str): The path to the YAML file to be opened and parsed.

    Returns:
        dict: The parsed contents of the YAML file.

    See also:
        _open_yaml: The function used to parse the YAML contents.
    """
    with codecs.open(yaml_file, "r", "utf8") as stream:
        yaml_contents = _open_yaml(stream, original_file=yaml_file)
    return yaml_contents


def ordered_load(stream, Loader=yaml_Loader, object_pairs_hook=OrderedDict):
    """
    Load a YAML stream while preserving the order of dictionaries.

    This function is a drop-in replacement for `yaml.load()`, but it ensures that the order of
    dictionaries is preserved by using `OrderedDict`.

    Args:
        stream (str): The YAML stream to load.
        Loader (yaml.Loader, optional): The YAML loader class to use. Defaults to `yaml_Loader`.
        object_pairs_hook (callable, optional): A callable that will be used to construct
                                                ordered mappings. Defaults to `OrderedDict`.

    Returns:
        OrderedDict: The loaded YAML data with preserved order of dictionaries.

    Raises:
        yaml.scanner.ScannerError: If there is an error when trying to load the stream.

    Notes:
        - If there is a `ScannerError`, additional hints will be printed to help diagnose
          common issues such as unquoted colons or other special symbols in the stream.
        - The function will exit the program with status code 1 if a `ScannerError` occurs.
    """
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        """
        Constructs a mapping from a YAML node.

        This function flattens the mapping of the given node and then constructs pairs from the
        node using the loader's `construct_pairs` method. The resulting pairs are then processed
        by the `object_pairs_hook`.

        Args:
            loader: The YAML loader instance.
            node: The YAML node to construct the mapping from.

        Returns:
            The constructed mapping, processed by `object_pairs_hook`.
        """
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    try:
        yaml_loaded = yaml.load(stream, OrderedLoader)
    except yaml.scanner.ScannerError as e:
        print("Error when trying to load the stream: {}".format(e))
        print("HINT: Ansible tasks names may cause expansion errors if not properly quoted.")
        print("HINT: Check for unquoted colons or other special symbols in the stream:")
        print("{}".format(stream))
        sys.exit(1)
    return yaml_loaded


def ordered_dump(data, stream=None, Dumper=yaml_Dumper, **kwds):
    """
    Serializes a Python object into a YAML stream, preserving the order of dictionaries.

    This function is a drop-in replacement for `yaml.dump()`, but it ensures that the order of
    dictionaries is preserved. It also includes custom representers for dictionaries and strings,
    and applies specific formatting fixes for YAML output.

    Args:
        data (Any): The Python object to serialize.
        stream (Optional[IO]): The stream to write the YAML output to. If None, the YAML string is
                               returned.
        Dumper (yaml.Dumper): The YAML dumper class to use. Defaults to `yaml_Dumper`.
        **kwds: Additional keyword arguments to pass to `yaml.dump()`.

    Returns:
        Optional[str]: The YAML string if `stream` is None, otherwise None.
    """
    class OrderedDumper(Dumper):
        """
        OrderedDumper is a custom YAML Dumper class that ensures the correct indentation of tags
        when dumping YAML data. It inherits from the base Dumper class and overrides the
        increase_indent method to fix tag indentations.
        """
        # fix tag indentations
        def increase_indent(self, flow=False, indentless=False):
            """
            Increases the indentation level for the YAML dumper.

            Args:
                flow (bool): Indicates if the flow style is used. Defaults to False.
                indentless (bool): Indicates if the indentation should be suppressed.
                                   Defaults to False.

            Returns:
                The result of the superclass's increase_indent method with the given flow and
                False for indentless.
            """
            return super(OrderedDumper, self).increase_indent(flow, False)

    def _dict_representer(dumper, data):
        """
        Custom YAML representer for Python dictionaries.

        This function is used to define how Python dictionaries should be represented when
        converting to YAML format. It uses the default mapping tag from the YAML resolver.

        Args:
            dumper (yaml.Dumper): The YAML dumper instance.
            data (dict): The Python dictionary to represent.

        Returns:
            yaml.Node: The YAML node representing the dictionary.
        """
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    def _str_representer(dumper, data):
        """
        Custom YAML representer for strings.

        This function is used to represent strings in YAML format. If the string contains a
        newline character, it will be represented as a block scalar (using the '|' style).
        Otherwise, it will be represented as a plain scalar.

        Args:
            dumper (yaml.Dumper): The YAML dumper instance.
            data (str): The string data to be represented.

        Returns:
            yaml.Node: The YAML node representing the string.
        """
        if '\n' in data:
            return dumper.represent_scalar(u'tag:yaml.org,2002:str', data,
                                           style='|')
        else:
            return dumper.represent_str(data)

    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    OrderedDumper.add_representer(str, _str_representer)

    # Fix formatting by adding a space in between tasks
    unformatted_yaml = yaml.dump(data, None, OrderedDumper, **kwds)
    formatted_yaml = re.sub(r"[\n]+([\s]*)- name", r"\n\n\1- name", unformatted_yaml)

    # Fix CDumper issue where it adds yaml document ending '...'
    # in some templated ansible remediations
    formatted_yaml = re.sub(r"\n\s*\.\.\.\s*", r"\n", formatted_yaml)

    if stream is not None:
        return stream.write(formatted_yaml)
    else:
        return formatted_yaml


def _strings_to_list(one_or_more_strings):
    """
    Convert a string or an iterable of strings into a list of strings.

    Args:
        one_or_more_strings (str or iterable): A single string or an iterable of strings.

    Returns:
        list: A list containing the input string or the elements of the input iterable.

    Examples:
        >>> _strings_to_list("hello")
        ['hello']
        >>> _strings_to_list(["hello", "world"])
        ['hello', 'world']
    """
    if isinstance(one_or_more_strings, str):
        return [one_or_more_strings]
    else:
        return list(one_or_more_strings)


def update_yaml_list_or_string(current_contents, new_contents, prepend=False):
    """
    Update a YAML list or string by combining current and new contents.

    This function takes the current contents and new contents, both of which can be either
    a string or a list of strings, and combines them into a single list or string. If the
    `prepend` flag is set to True, the new contents are added before the current contents.
    Otherwise, the new contents are appended to the current contents.

    Args:
        current_contents (str or list of str): The existing contents to be updated.
        new_contents (str or list of str): The new contents to be added.
        prepend (bool): If True, new contents are added before current contents. Defaults to False.

    Returns:
        str or list of str: The updated contents. If the result is a single item, it is returned
                            as a string. If the result is empty, an empty string is returned.
    """
    result = []
    if current_contents:
        result += _strings_to_list(current_contents)
    if new_contents:
        if prepend:
            result = _strings_to_list(new_contents) + result
        else:
            result += _strings_to_list(new_contents)
    if not result:
        result = ""
    if len(result) == 1:
        result = result[0]
    return result


def convert_string_to_bool(string):
    """
    Converts a string representation of a boolean to its corresponding boolean value.

    Args:
        string (str): The string to convert. Expected values are "true" or "false"
                      (case insensitive).

    Returns:
        bool: True if the string is "true" (case insensitive), False if the string is "false"
              (case insensitive).

    Raises:
        ValueError: If the string is not "true" or "false" (case insensitive).
    """
    lower = string.lower()
    if lower == "true":
        return True
    elif lower == "false":
        return False
    else:
        raise ValueError(
                    "Invalid value %s while expecting boolean string" % string)



from __future__ import annotations

from openpyxl.worksheet.worksheet import Worksheet

# The start row is 2, to avoid importing the header
START_ROW = 2


class Row:
    row_id = 0
    IA_Control = ""
    CCI = ""
    SRGID = ""
    STIGID = ""
    SRG_Requirement = ""
    Requirement = ""
    SRG_VulDiscussion = ""
    Vul_Discussion = ""
    Status = ""
    SRG_Check = ""
    Check = ""
    SRG_Fix = ""
    Fix = ""
    Severity = ""
    Mitigation = ""
    Artifact_Description = ""
    Status_Justification = ""

    @staticmethod
    def from_row(sheet: Worksheet, row_number: int, full_name: str, changed_name: str) -> Row:
        self = Row()
        self.row_id = row_number
        self.IA_Control = sheet[f'A{row_number}'].value
        self.CCI = sheet[f'B{row_number}'].value
        self.SRGID = sheet[f'C{row_number}'].value
        self.STIGID = sheet[f'D{row_number}'].value
        self.SRG_Requirement = sheet[f'E{row_number}'].value
        if sheet[f'F{row_number}'].value:
            self.Requirement = sheet[f'F{row_number}'].value.replace(full_name, changed_name)
        self.SRG_VulDiscussion = sheet[f'G{row_number}'].value
        if sheet[f'H{row_number}'].value:
            self.Vul_Discussion = sheet[f'H{row_number}'].value.replace(full_name, changed_name)
        self.Status = sheet[f'I{row_number}'].value
        self.SRG_Check = sheet[f'J{row_number}'].value
        if sheet[f'K{row_number}'].value:
            self.Check = sheet[f'K{row_number}'].value.replace(full_name, changed_name)
        self.SRG_Fix = sheet[f'L{row_number}'].value
        if sheet[f'M{row_number}'].value:
            self.Fix = sheet[f'M{row_number}'].value.replace(full_name, changed_name)
        self.Severity = sheet[f'N{row_number}'].value
        self.Mitigation = sheet[f'O{row_number}'].value
        self.Artifact_Description = sheet[f'P{row_number}'].value
        self.Artifact_Description = sheet[f'Q{row_number}'].value
        return self

    def __str__(self) -> str:
        return f'<Row {self.row_id}>'


def get_stigid_set(sheet: Worksheet, end_row: int) -> set[str]:
    result = set()
    for i in range(START_ROW, end_row):
        cci_raw = sheet[f'D{i}'].value

        if cci_raw is None or cci_raw == "":
            continue
        result.add(cci_raw)
    return result


def get_cce_dict_to_row_dict(sheet: Worksheet, full_name: str, changed_name: str,
                             end_row: int) -> dict:
    result = dict()
    for i in range(START_ROW, end_row):
        cci_raw = sheet[f'D{i}'].value
        if cci_raw is None or cci_raw == "":
            continue
        cci = cci_raw.strip().replace('\n', '')
        result[cci] = Row.from_row(sheet, i, full_name, changed_name)

    return result



from __future__ import print_function
import re
import json


def normalize_policy_path(path_string):
    """Generate a normalized path array from yaml.
       Paths are space-separated key names.
       All firefox policies are under the policies key.
    """
    path = []
    if len(path_string.strip()) > 0:
        path = re.split(" ", path_string)
    if path[0] != 'policies':
        path = ['policies'] + path
    return path


def make_json_boolean_match_regex(json_value):
    if json_value.lower() in ["false", "true"]:
        return r"[{0}{1}]{2}".format(json_value[0].upper(),
                                     json_value[0].lower(),
                                     json_value[1:]
                                     )
    return "\"{0}\"".format(json_value)


def make_json_boolean_remediate(json_value):
    if json_value.lower() in ["false", "true"]:
        return json_value.lower() == "true"
    return "\'{0}\'".format(json_value)


# These regexes work under the complexities currently observed with the Firefox
# policies.json, which is (so far) a subset of JSON that is just nested keys.
# No JSON arrays are supported.
# Remove the moment that OpenSCAP/OVAL spec supports a JSON configuration probe.
OVAL_MATCH_ANY_FOLLOWING = r"[^}]*\}"
OVAL_MATCH_OPEN_BRACES_FOR_BLOCK = r"\{[\s\S]*?"
OVAL_MATCH_ANYTHING_AFTER_OPEN_BRACE = r"\{[\s\S]*"

# Note: str.format() notation.
OVAL_MATCH_JSON_KEY = r'(?="{0}")"{0}"[\s]*:[\s]*'


def regex_escape(_search_string):
    return re.escape(_search_string)


def build_oval_search_regex_for_json_parameter(_path, parameter):
    parameter = r'(?=[^"]){0}"[\s]*:[\s]*([^,}}]+),?'.format(parameter)
    _result = parameter
    for p in reversed(_path):
        depth_match_block = OVAL_MATCH_OPEN_BRACES_FOR_BLOCK
        if p == "policies":
            depth_match_block = OVAL_MATCH_ANYTHING_AFTER_OPEN_BRACE

        _result = ''.join([OVAL_MATCH_JSON_KEY.format(regex_escape(p)),
                           depth_match_block,
                           _result])
    _result = r'^(?i)\{\s*' + _result + r'\s*'
    return _result


def build_python_json_notation(_path, depth=0):
    """Generate the Python call syntax for this path within a given distance
       from the top-level node.

       Parameters:
           _path: the list of the entire path in the JSON document to this
           particular key.
           depth: How far along the path to generate this syntax.
    """
    if depth > len(_path):
        return r''
    _result = r''
    for p in list(reversed(_path))[depth:]:
        _result = ''.join(["['{0}']".format(p), _result])
    return _result


def build_test_json(policy, _test, missing=False, wrong=False):
    """Generate a minimal JSON configuration for this particular policy object, based
       on a correct (set to value), missing, or wrong value.
       Returns a dict() containing this test entry.

       Parameters:
           policy: current policy item to generate for
           _test: dict() that has at least the value {"policies": {} }
    """
    _current = _test.get("policies")
    _key = policy["parameter"]
    _value = make_json_boolean_remediate(policy["value"])

    for p in policy["subpath"]:
        if p not in _current:
            _current[p] = dict()
        _current = _current.get(p)

    if wrong:
        _current[_key] = "VERYVERYBADVALUE"
    elif not missing:
        _current[_key] = _value
    return _test


def preprocess(data, lang):
    """Generate the appropriate regex for each policy path, search values, etc.
       Expects a dict of template arguments:
           path - Space-separated path in JSON to dictate where the parameter is. "policies" is
                  added to the front if not present or used if path is not present.
           parameter - Parameter name to search for the state of the value.
           value - Expected value to use for remediation.
           search_value (optional) - custom regular expression to use in the match for value.
    """
    _test_correct_config = {"policies": {}}
    _test_bad_missing = {"policies": {}}
    _test_bad_wrong = {"policies": {}}
    for i, _policy in enumerate(data.get("policies", [])):
        _path = normalize_policy_path(_policy.get("path", ""))

        # regex used in OVAL for detection
        _policy["oval_regex"] = build_oval_search_regex_for_json_parameter(_path,
                                                                           _policy["parameter"])
        _policy["oval_state"] = make_json_boolean_match_regex(_policy["value"])
        # Implement override for search_value
        if _policy.get("search_value", None):
            _policy["oval_state"] = _policy["search_value"]

        # path without "policies"
        _policy["subpath"] = _path[1:]
        # Precached notation for addressing JSON path
        _policy["path_python"] = [build_python_json_notation(_path, x)
                                  for x in range(0, len(_path) + 1)][:-1]
        # JSON path as iterable list, sans "policies"
        _policy["subpath_string"] = '_'.join(_path[1:])
        # entire JSON path to be set as an iterable list.
        _policy["path_list"] = _path
        # value with quotes added for simple insertion.
        _policy["value_escaped"] = make_json_boolean_remediate(_policy["value"])

        # correct_value.pass tests
        build_test_json(_policy,
                        _test_correct_config,
                        missing=False,
                        wrong=False)

        # missing_value.fail tests
        build_test_json(_policy,
                        _test_bad_missing,
                        missing=True,
                        wrong=False)

        # wrong_value.fail tests
        build_test_json(_policy,
                        _test_bad_wrong,
                        missing=False,
                        wrong=True)

        # Reassign the modified policy entry.
        data["policies"][i] = _policy

    # Pretty-print the output for our test JSON to variables
    # so that it can be used in the test templates.
    data["_correct_config"] = json.dumps(_test_correct_config, indent=4, separators=(',', ': '))
    data["_bad_missing"] = json.dumps(_test_bad_missing, indent=4, separators=(',', ': '))
    data["_bad_wrong"] = json.dumps(_test_bad_wrong, indent=4, separators=(',', ': '))
    return data



import json
import os

import pytest

from ssg_test_suite import common
import analyze_results as analyze


@pytest.fixture
def raw_results():
    return json.load(open(os.path.join(os.path.dirname(__file__), "data", "rules.json"), "r"))


@pytest.fixture
def results_list(raw_results):
    return [common.RuleResult(r) for r in raw_results.values()]


@pytest.fixture
def different_results(raw_results):
    return [common.RuleResult(raw_results[key])
            for key in ("container_failed_remediation", "vm_passed_everything")]


def test_success(raw_results):
    rule_ok = common.RuleResult(raw_results["vm_passed_everything"])
    assert rule_ok.success

    rule_not_ok = common.RuleResult(raw_results["container_failed_remediation"])
    assert not rule_not_ok.success


def test_aggregation(results_list):
    aggregated = analyze.aggregate_results_by_scenarios(results_list)
    assert len(aggregated) == 3


def test_differences(different_results):
    assert sorted(different_results)[0].success

    difference = analyze.analyze_differences(different_results)
    assert difference.why_failed == "remediation"


def test_persistence(raw_results):
    for r in raw_results.values():
        intermediate = common.RuleResult(r)
        assert r == intermediate.save_to_dict()



"""
Common functions for processing Checks in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import os

from .constants import XCCDF12_NS
from .oval import parse_affected
from .utils import read_file_list


def get_content_ref_if_exists_and_not_remote(check):
    """
    Given an OVAL check element, examine the ``xccdf_ns:check-content-ref``.

    If the check-content-ref element exists and it isn't remote, return it.

    Args:
        check (Element): An OVAL check element to be examined.

    Returns:
        Element or None: The check-content-ref element if it exists and is not remote, otherwise None.

    See Also:
        is_content_href_remote: Function to determine if the content reference is remote.
    """
    checkcontentref = check.find("./{%s}check-content-ref" % XCCDF12_NS)
    if checkcontentref is None:
        return None
    if is_content_href_remote(checkcontentref):
        return None
    return checkcontentref


def is_content_href_remote(check_content_ref):
    """
    Given an OVAL check-content-ref element, examine the 'href' attribute.

    Args:
        check_content_ref (Element): An XML element representing the OVAL check-content-ref.

    Returns:
        bool: True if the 'href' attribute starts with 'http://' or 'https://', otherwise False.

    Raises:
        RuntimeError: If the 'href' attribute does not exist in the check_content_ref element.
    """
    hrefattr = check_content_ref.get("href")
    if hrefattr is None:
        # @href attribute of <check-content-ref> is required by XCCDF standard
        msg = "Invalid OVAL <check-content-ref> detected - missing the " \
              "'href' attribute!"
        raise RuntimeError(msg)

    return hrefattr.startswith("http://") or hrefattr.startswith("https://")


def get_oval_path(rule_obj, oval_id):
    """
    Returns the full path to the OVAL check file for the given rule object and OVAL ID.

    Args:
        rule_obj (dict): A dictionary containing rule information.
                         It must include the keys 'dir', 'id', and 'ovals'.
        oval_id (str): A string representing the ID of the OVAL check file.
                       If it does not end with ".xml", the extension will be appended.

    Returns:
        str: The full path to the OVAL check file.

    Raises:
        ValueError: If the rule_obj is malformed or if the oval_id is unknown for the given rule.
    """
    if not oval_id.endswith(".xml"):
        oval_id += ".xml"

    if 'dir' not in rule_obj or 'id' not in rule_obj:
        raise ValueError("Malformed rule_obj")

    if 'ovals' not in rule_obj or oval_id not in rule_obj['ovals']:
        raise ValueError("Unknown oval_id:%s for rule_id" % oval_id)

    return os.path.join(rule_obj['dir'], 'oval', oval_id)


def get_oval_contents(rule_obj, oval_id):
    """
    Returns the tuple (path, contents) of the check described by the given oval_id or product.

    Parameters:
        rule_obj (object): The rule object containing the OVAL definitions.
        oval_id (str): The identifier of the OVAL check.

    Returns:
        tuple: A tuple containing the path to the OVAL file and its contents.
    """
    oval_path = get_oval_path(rule_obj, oval_id)
    oval_contents = read_file_list(oval_path)

    return oval_path, oval_contents


def set_applicable_platforms(oval_contents, new_platforms):
    """
    Modifies the given OVAL contents to update the platforms to the new platforms.

    Args:
        oval_contents (list of str): The original OVAL content lines.
        new_platforms (list of str): The new platforms to be set in the OVAL content.

    Returns:
        list of str: The modified OVAL content lines with updated platforms.
    """
    start_affected, end_affected, indent = parse_affected(oval_contents)

    platforms = sorted(new_platforms)
    new_platforms_xml = map(lambda x: indent + "<platform>%s</platform>" % x, platforms)
    new_platforms_xml = list(new_platforms_xml)

    new_contents = oval_contents[0:start_affected+1]
    new_contents.extend(new_platforms_xml)
    new_contents.extend(oval_contents[end_affected:])

    return new_contents



def preprocess(data, lang):
    if lang == "oval" and data["attribute"] == 'permissions':
        # create STATEMODE used in the OVAL template by processing the octal permission and
        # creating the equivalent permission fields of "unix:file_state" element.
        mode = data["value"]
        fields = [
            'oexec', 'owrite', 'oread', 'gexec', 'gwrite', 'gread',
            'uexec', 'uwrite', 'uread', 'sticky', 'sgid', 'suid']
        mode_int = int(mode, 8)
        mode_str = ""
        for field in fields:
            if mode_int & 0x01 == 0:
                mode_str = (
                    "<unix:{field} datatype=\"boolean\">false</unix:{field}>\n{mode_str}".format(
                        field=field, mode_str=mode_str))
            mode_int = mode_int >> 1
        data["statemode"] = mode_str.rstrip("\n")
    return data



import os
import re

from ssg.rule_yaml import find_section_lines, get_yaml_contents
from ssg.utils import read_file_list
from . import create_output, write_output


def add_replacement_to_result(replacement, result):
    for line in replacement.split("\n"):
        if line != "":
            result = (*result, f"    {line}",)
        else:
            result = (*result, "")
    return result


def replace_yaml_section(section: str, replacement: str, rule_dir: dict) -> None:
    path = create_output(rule_dir['dir'])

    lines = read_file_list(path)
    replacement = replacement.replace('<', '&lt;').replace('>', '&gt;')
    replacement = re.sub(r'Satisfies: SRG-OS.+', '', replacement)
    section_ranges = find_section_lines(lines, section)
    if section_ranges:
        result = lines[:section_ranges[0].start]
        result = (*result, f"{section}: |-")
        result = add_replacement_to_result(replacement, result)
        end_line = section_ranges[0].end
        for line in lines[end_line:]:
            result = [*result, line]
        result = [*result, '\n']
    else:
        result = lines
        result = (*result, f"\n{section}: |-")
        result = add_replacement_to_result(replacement, result)

    write_output(path, result)


def replace_yaml_key(key: str, replacement: str, rule_dir: dict) -> None:
    path_dir = rule_dir['dir']
    path = os.path.join(path_dir, 'rule.yml')
    lines = get_yaml_contents(rule_dir)
    section_ranges = find_section_lines(lines.contents, key)
    replacement_line = f"{key}: {replacement}"
    if section_ranges:
        result = lines.contents[:section_ranges[0].start]
        result = (*result, replacement_line)
        end_line = section_ranges[0].end
        for line in lines.contents[end_line:]:
            result = (*result, line)
    else:
        result = lines.contents
        result = (*result, replacement_line)

    with open(path, 'w') as f:
        for line in result:
            f.write(line.rstrip())


def update_row(changed: str, current: str, rule_dir_json: dict, section: str) -> None:
    if changed != current and changed:
        replace_yaml_section(section, changed, rule_dir_json)



"""
Utils functions consumed by SSG
"""

from __future__ import absolute_import
from __future__ import print_function

import multiprocessing
import errno
import os
import re
from collections import namedtuple
import hashlib

from .constants import (FULL_NAME_TO_PRODUCT_MAPPING,
                        MAKEFILE_ID_TO_PRODUCT_MAP,
                        MULTI_PLATFORM_LIST,
                        )


class SSGError(RuntimeError):
    pass


PRODUCT_NAME_PARSER = re.compile(r"([a-zA-Z\-]+)([0-9]+)")


class VersionSpecifierSet(set):
    """
    A set-like collection that only accepts VersionSpecifier objects.

    This class extends the built-in set and ensures that all elements are instances of the
    VersionSpecifier class. It provides additional properties to generate titles, CPE IDs, and
    OVAL IDs from the contained VersionSpecifier objects.

    Attributes:
        title (str): A string representation of the set, joining the titles of the contained
                     VersionSpecifier objects with ' and '.
        cpe_id (str): A string representation of the set, joining the CPE IDs of the contained
                      VersionSpecifier objects with ':'.
        oval_id (str): A string representation of the set, joining the OVAL IDs of the contained
                       VersionSpecifier objects with '_'.
    """
    def __init__(self, s=()):
        for el in s:
            if not isinstance(el, VersionSpecifier):
                raise ValueError('VersionSpecifierSet can only work with VersionSpecifier objects,'
                                 ' invalid object: {0}'.format(repr(el)))
        super(VersionSpecifierSet, self).__init__(s)

    @property
    def title(self):
        return ' and '.join([ver_spec.title for ver_spec in sorted(self)])

    @property
    def cpe_id(self):
        return ':'.join([ver_spec.cpe_id for ver_spec in sorted(self)])

    @property
    def oval_id(self):
        return '_'.join([ver_spec.oval_id for ver_spec in sorted(self)])


class VersionSpecifier:
    """
    A class to represent a version specifier with properties and methods to manipulate and
    retrieve version information.

    Attributes:
        op (str): The operation associated with the version specifier.
        _evr_ver_dict (dict): A dictionary containing epoch, version, and release information.
    """
    def __init__(self, op, evr_ver_dict):
        self._evr_ver_dict = evr_ver_dict
        self.op = op

    def __str__(self):
        return '{0} {1}'.format(self.op, self.ver)

    def __repr__(self):
        return '<VersionSpecifier({0},{1})>'.format(self.op, self.ver)

    def __hash__(self):
        return hash(self.op + self.ver)

    def __eq__(self, other):
        return self.op+self.ver == other.op+other.ver

    def __lt__(self, other):
        return self.op+self.ver < other.op+other.ver

    @property
    def evr_op(self):
        return comparison_to_oval(self.op)

    @property
    def ver(self):
        return VersionSpecifier.evr_dict_to_str(self._evr_ver_dict)

    @property
    def evr_ver(self):
        return VersionSpecifier.evr_dict_to_str(self._evr_ver_dict, True)

    @property
    def ev_ver(self):
        return VersionSpecifier.evr_dict_to_str(self._evr_ver_dict, True).split("-")[0]

    @property
    def title(self):
        return '{0} {1}'.format(comparison_to_oval(self.op), self.ver)

    @property
    def cpe_id(self):
        return '{0}:{1}'.format(escape_comparison(self.op), self.ver)

    @property
    def oval_id(self):
        return '{0}_{1}'.format(escape_comparison(self.op), escape_id(self.ver))

    @staticmethod
    def evr_dict_to_str(evr, fully_formed_evr_string=False):
        """
        Convert an EVR (Epoch-Version-Release) dictionary to a string representation.

        Args:
            evr (dict): A dictionary containing 'epoch', 'version', and 'release' keys.
                        'epoch' and 'release' can be None.
            fully_formed_evr_string (bool): If True, ensures that the returned string includes '0'
                                            for missing 'epoch' and '-0' for missing 'release'.

        Returns:
            str: The string representation of the EVR dictionary.
        """
        res = ''
        if evr['epoch'] is not None:
            res += evr['epoch'] + ':'
        elif fully_formed_evr_string:
            res += '0:'
        res += evr['version']
        if evr['release'] is not None:
            res += '-' + evr['release']
        elif fully_formed_evr_string:
            res += '-0'
        return res


def map_name(version):
    """
    Maps SSG Makefile internal product name to official product name.

    This function takes a version string and maps it to an official product name based on
    predefined mappings. It handles multi-platform versions by trimming the "multi_platform_"
    prefix and recursively mapping the trimmed version.

    Args:
        version (str): The version string to be mapped.

    Returns:
        str: The official product name corresponding to the given version.

    Raises:
        RuntimeError: If the version is invalid or cannot be mapped to any known product.
    """
    if version.startswith("multi_platform_"):
        trimmed_version = version[len("multi_platform_"):]
        if trimmed_version not in MULTI_PLATFORM_LIST:
            raise RuntimeError(
                "%s is an invalid product version. If it's multi_platform the "
                "suffix has to be from (%s)."
                % (version, ", ".join(MULTI_PLATFORM_LIST))
            )
        return map_name(trimmed_version)

    # By sorting in reversed order, keys which are a longer version of other keys are
    # visited first (e.g., rhosp vs. rhel)
    for key in sorted(MAKEFILE_ID_TO_PRODUCT_MAP, reverse=True):
        if version.startswith(key):
            return MAKEFILE_ID_TO_PRODUCT_MAP[key]

    raise RuntimeError("Can't map version '%s' to any known product!"
                       % (version))


def product_to_name(prod):
    """
    Converts a product identifier into a full product name.

    This function takes a product identifier and attempts to match it with a full product name
    from the `FULL_NAME_TO_PRODUCT_MAPPING` dictionary. If the product identifier is found in
    the dictionary, the corresponding full product name is returned. If the product identifier
    is in the `MULTI_PLATFORM_LIST` or is 'all', a string prefixed with "multi_platform_" is
    returned. If the product identifier is not recognized, a RuntimeError is raised.

    Args:
        prod (str): The product identifier to convert.

    Returns:
        str: The full product name corresponding to the given product identifier.

    Raises:
        RuntimeError: If the product identifier is not recognized.
    """
    for name, prod_type in FULL_NAME_TO_PRODUCT_MAPPING.items():
        if prod == prod_type:
            return name
    if prod in MULTI_PLATFORM_LIST or prod == 'all':
        return "multi_platform_" + prod
    raise RuntimeError("Unknown product name: %s" % prod)


def name_to_platform(names):
    """
    Converts one or more full names to a string containing one or more <platform> elements.

    Args:
        names (str or list of str): A single name as a string or a list of names.

    Returns:
        str: A string containing one or more <platform> elements.
    """
    if isinstance(names, str):
        return "<platform>%s</platform>" % names
    return "\n".join(map(name_to_platform, names))


def product_to_platform(prods):
    """
    Converts one or more product ids into a string with one or more <platform> elements.

    Args:
        prods (str or list of str): A single product id as a string or a list of product ids.

    Returns:
        str: A string containing one or more <platform> elements.
    """
    if isinstance(prods, str):
        return name_to_platform(product_to_name(prods))
    return "\n".join(map(product_to_platform, prods))


def parse_name(product):
    """
    Parses a given product string and returns a namedtuple containing the name and version.

    Args:
        product (str): The product string to parse, e.g., "rhel9".

    Returns:
        namedtuple: A namedtuple with 'name' and 'version' attributes, e.g., ("rhel", "9").

    Example:
        >>> parse_name("rhel9")
        product(name='rhel', version='9')
    """
    prod_tuple = namedtuple('product', ['name', 'version'])

    _product = product
    _product_version = None
    match = PRODUCT_NAME_PARSER.match(product)

    if match:
        _product = match.group(1)
        _product_version = match.group(2)

    return prod_tuple(_product, _product_version)


def parse_platform(platform):
    """
    Parses a comma-separated string of platforms and returns a set of trimmed platform names.

    Args:
        platform (str): A comma-separated string of platform names.

    Returns:
        set: A set of platform names with leading and trailing whitespace removed.
    """
    return set(map(lambda x: x.strip(), platform.split(',')))


def get_fixed_product_version(product, product_version):
    """
    Adjusts the product version format for specific products.

    Some product versions have a dot in between the numbers while the product ID doesn't have the
    dot, but the full product name does. This function ensures the correct format for the product
    version.

    Args:
        product (str): The name of the product (e.g., "ubuntu", "macos").
        product_version (str): The version of the product as a string.

    Returns:
        str: The adjusted product version with the correct format.
    """
    if product == "ubuntu" or product == "macos":
        product_version = product_version[:2] + "." + product_version[2:]
    return product_version


def is_applicable_for_product(platform, product):
    """
    Determines if a remediation script is applicable for a given product based on the platform specifier.

    The function checks if the platform is either a general multi-platform specifier or matches
    the specific product name and version.

    Args:
        platform (str): The platform specifier of the remediation script.
        product (str): The product name and version.

    Returns:
        bool: True if the remediation script is applicable for the product, False otherwise.
    """
    # If the platform is None, platform must not exist in the config, so return False.
    if not platform:
        return False

    product, product_version = parse_name(product)

    # Define general platforms
    multi_platforms = ['multi_platform_all',
                       'multi_platform_' + product]

    # First test if platform isn't for 'multi_platform_all' or 'multi_platform_' + product
    for _platform in multi_platforms:
        if _platform in platform and product in MULTI_PLATFORM_LIST:
            return True

    product_name = ""
    # Get official name for product
    if product_version is not None:
        product_name = map_name(product) + ' ' + get_fixed_product_version(
            product, product_version
        )
    else:
        product_name = map_name(product)

    # Test if this is for the concrete product version
    for _name_part in platform.split(','):
        if product_name == _name_part.strip():
            return True

    # Remediation script isn't neither a multi platform one, nor isn't
    # applicable for this product => return False to indicate that
    return False


def is_applicable(platform, product):
    """
    Check if a platform is applicable for the given product.

    This function determines whether a specified platform or a list of platforms is applicable
    for a given product. It handles cases where the platform is specified as 'all' or
    'multi_platform_all', as well as cases where the platform is a comma-separated list of
    products.

    Args:
        platform (str): The platform or list of platforms to check.
        product (str): The product to check applicability for.

    Returns:
        bool: True if the product is applicable for the platform or list of platforms, False otherwise.
    """
    if platform == 'all' or platform == 'multi_platform_all':
        return True

    if is_applicable_for_product(platform, product):
        return True

    if 'osp7' in product and 'osp7' in platform:
        return True

    return product in platform.split(',')


def required_key(_dict, _key):
    """
    Returns the value associated with the specified key in the given dictionary.

    Parameters:
        _dict (dict): The dictionary to search for the key.
        _key: The key to look for in the dictionary.

    Returns:
        The value associated with the specified key in the dictionary.

    Raises:
        ValueError: If the key is not found in the dictionary, an exception is raised with a
                    message indicating that the key is required but was not found.
    """
    if _key in _dict:
        return _dict[_key]

    raise ValueError("%s is required but was not found in:\n%s" %
                     (_key, repr(_dict)))


def get_cpu_count():
    """
    Returns the most likely estimate of the number of CPUs in the machine for threading purposes,
    gracefully handling errors and possible exceptions.

    Returns:
        int: The number of CPUs available. If the number of CPUs cannot be determined, returns 2
             as a default value.
    """
    try:
        return max(1, multiprocessing.cpu_count())

    except NotImplementedError:
        return 2


def merge_dicts(left, right):
    """
    Merges two dictionaries, keeping left and right as passed.

    If there are any common keys between left and right, the value from right is used.

    Args:
        left (dict): The first dictionary.
        right (dict): The second dictionary.

    Returns:
        dict: The merged dictionary containing keys and values from both left and right dictionaries.
    """
    result = left.copy()
    result.update(right)
    return result


def subset_dict(dictionary, keys):
    """
    Restricts a dictionary to only include specified keys.

    This function creates a new dictionary that contains only the key-value pairs from the
    original dictionary where the key is present in the provided list of keys. Neither the
    original dictionary nor the list of keys is modified.

    Args:
        dictionary (dict): The original dictionary from which to create the subset.
        keys (iterable): An iterable of keys that should be included in the subset.

    Returns:
        dict: A new dictionary containing only the key-value pairs where the key is in the
              provided list of keys.
    """
    result = dictionary.copy()
    for original_key in dictionary:
        if original_key not in keys:
            del result[original_key]

    return result


def read_file_list(path):
    """
    Reads the given file path and returns the contents as a list.

    Args:
        path (str): The path to the file to be read.

    Returns:
        list: A list containing the contents of the file, split by lines or other delimiters as
              defined by `split_string_content`.

    Raises:
        FileNotFoundError: If the file at the given path does not exist.
        IOError: If an I/O error occurs while reading the file.
    """
    with open(path, 'r') as f:
        return split_string_content(f.read())


def split_string_content(content):
    """
    Splits the input string content by newline characters and returns the result as a list of strings.

    Args:
        content (str): The string content to be split.

    Returns:
        list: A list of strings, each representing a line from the input content.
              The last element will be removed if it is an empty string.
    """
    file_contents = content.split("\n")
    if file_contents[-1] == '':
        file_contents = file_contents[:-1]
    return file_contents


def write_list_file(path, contents):
    """
    Writes the given contents to the specified file path.

    Args:
        path (str): The file path where the contents will be written.
        contents (list of str): A list of strings to be written to the file.

    Returns:
        None
    """
    _contents = "\n".join(contents) + "\n"
    _f = open(path, 'w')
    _f.write(_contents)
    _f.flush()
    _f.close()


# Taken from https://stackoverflow.com/a/600612/592892
def mkdir_p(path):
    """
    Create a directory and all intermediate-level directories if they do not exist.

    Args:
        path (str): The directory path to create.

    Returns:
        bool: True if the directory was created, False if it already exists.

    Raises:
        OSError: If the directory cannot be created and it does not already exist.
    """
    if os.path.isdir(path):
        return False
    # Python >=3.4.1
    # os.makedirs(path, exist_ok=True)
    try:
        os.makedirs(path)
        return True
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            return False
        else:
            raise


def escape_regex(text):
    """
    Escapes special characters in a given text to make it safe for use in regular expressions.

    This function mimics the behavior of re.escape() in Python 3.7, which escapes a reasonable set
    of characters.
    Specifically, it escapes the following characters: #, $, &, *, +, ., ^, `, |, ~, :, (, ), and -.
    Note that the characters '!', '"', '%', "'", ',', '/', ':', ';', '<', '=', '>', '@', and "`"
    are not escaped.

    Args:
        text (str): The input string containing characters to be escaped.

    Returns:
        str: A new string with special characters escaped.
    """
    # We could use re.escape(), but it escapes too many characters, including plain white space.
    # In python 3.7 the set of characters escaped by re.escape is reasonable, so lets mimic it.
    # See https://docs.python.org/3/library/re.html#re.sub
    # '!', '"', '%', "'", ',', '/', ':', ';', '<', '=', '>', '@', and "`" are not escaped.
    return re.sub(r"([#$&*+.^`|~:()\[\]-])", r"\\\1", text)


def escape_id(text):
    """
    Converts a given text into a string that is more readable and compatible with OSCAP/XCCDF/OVAL
    entity IDs by replacing non-word characters with underscores and stripping leading/trailing
    underscores.

    Args:
        text (str): The input string to be converted.

    Returns:
        str: The converted string with non-word characters replaced by underscores and
             leading/trailing underscores removed.
    """
    # Make a string used as an Id for OSCAP/XCCDF/OVAL entities more readable
    # and compatible with:
    # OVAL: r'oval:[A-Za-z0-9_\-\.]+:ste:[1-9][0-9]*'
    return re.sub(r"[^\w]+", "_", text).strip("_")


def escape_yaml_key(text):
    """
    Escapes uppercase letters and caret symbols in a YAML key by prefixing them with a caret
    and converts the entire key to lowercase.

    This function is used to handle the limitation of OVAL's name argument of the field type,
    which requires avoiding uppercase letters for keys. The probe would escape them with the
    '^' symbol.

    Example:
        myCamelCase^Key -> my^camel^case^^^key

    Args:
        text (str): The YAML key to be escaped.

    Returns:
        str: The escaped and lowercased YAML key.
    """
    return re.sub(r'([A-Z^])', '^\\1', text).lower()


def _map_comparison_op(op, table):
    """
    Maps a comparison operator to its corresponding function or value from a given table.

    Args:
        op (str): The comparison operator to be mapped.
        table (dict): A dictionary where keys are comparison operators and values are the
                      corresponding functions or values.

    Returns:
        The function or value corresponding to the given comparison operator.

    Raises:
        KeyError: If the given comparison operator is not found in the table.
    """
    if op not in table:
        raise KeyError("Invalid comparison operator: %s (expected one of: %s)",
                       op, ', '.join(table.keys()))
    return table[op]


def escape_comparison(op):
    """
    Maps a comparison operator to its corresponding string representation.

    Args:
        op (str): The comparison operator to be mapped.
                  Expected values are '==', '!=', '>', '<', '>=', '<='.

    Returns:
        str: The string representation of the comparison operator.
             Possible return values are 'eq', 'ne', 'gt', 'le', 'gt_or_eq', 'le_or_eq'.

    Raises:
        KeyError: If the provided operator is not in the mapping dictionary.
    """
    return _map_comparison_op(op, {
        '==': 'eq',       '!=': 'ne',
        '>': 'gt',        '<': 'le',
        '>=': 'gt_or_eq', '<=': 'le_or_eq',
    })


def comparison_to_oval(op):
    """
    Converts a comparison operator to its corresponding OVAL string representation.

    Args:
        op (str): The comparison operator to convert.
                  Expected values are '==', '!=', '>', '<', '>=', '<='.

    Returns:
        str: The OVAL string representation of the comparison operator.

    Raises:
        KeyError: If the provided operator is not one of the expected values.
    """
    return _map_comparison_op(op, {
        '==': 'equals',                '!=': 'not equal',
        '>': 'greater than',           '<': 'less than',
        '>=': 'greater than or equal', '<=': 'less than or equal',
    })


def sha256(text):
    """
    Generate a SHA-256 hash for the given text.

    Args:
        text (str): The input text to be hashed.

    Returns:
        str: The SHA-256 hash of the input text as a hexadecimal string.
    """
    return hashlib.sha256(text.encode('utf-8')).hexdigest()


def banner_regexify(banner_text):
    """
    Converts a banner text into a regex pattern.

    This function escapes special regex characters in the input text and then performs the
    following transformations:
    - Replaces newline characters ("\n") with a placeholder string "BFLMPSVZ".
    - Replaces spaces with a regex pattern that matches one or more whitespace characters or
      newline characters.
    - Replaces the placeholder string "BFLMPSVZ" with a regex pattern that matches one or more
      newline characters or escaped newline sequences.

    Args:
        banner_text (str): The banner text to be converted into a regex pattern.

    Returns:
        str: The resulting regex pattern.
    """
    return escape_regex(banner_text) \
        .replace("\n", "BFLMPSVZ") \
        .replace(" ", "[\\s\\n]+") \
        .replace("BFLMPSVZ", "(?:[\\n]+|(?:\\\\n)+)")


def banner_anchor_wrap(banner_text):
    """
    Wraps the given banner text with '^' at the beginning and '$' at the end.

    Args:
        banner_text (str): The text to be wrapped.

    Returns:
        str: The banner text wrapped with '^' at the beginning and '$' at the end.
    """
    return "^" + banner_text + "$"


def parse_template_boolean_value(data, parameter, default_value):
    """
    Parses a boolean value from a template parameter.

    Args:
        data (dict): The dictionary containing the template data.
        parameter (str): The key for the parameter to parse.
        default_value (bool): The default value to return if the parameter is not found or is None.

    Returns:
        bool: The parsed boolean value.

    Raises:
        ValueError: If the parameter value is not "true" or "false".
    """
    value = data.get(parameter)
    if not value:
        return default_value
    if value == "true":
        return True
    elif value == "false":
        return False
    else:
        raise ValueError(
            "Template parameter {} used in rule {} cannot accept the "
            "value {}".format(parameter, data["_rule_id"], value))


def check_conflict_regex_directory(data):
    """
    Validate that either all paths are directories or the 'file_regex' key exists.

    This function checks the following conditions:
    1. If the 'is_directory' key is present in the data dictionary, it ensures that all file paths
       in the 'filepath' list are either directories or files. Mixing directories and files is not
       allowed.
    2. If the 'file_regex' key is present in the data dictionary, it ensures that all file paths
       in the 'filepath' list are directories.

    Args:
        data (dict): A dictionary containing the following keys:
            - 'filepath' (list): A list of file paths to be validated.
            - 'is_directory' (bool, optional): A flag indicating whether the paths are directories.
            - 'file_regex' (str, optional): A regular expression pattern for file matching.
            - '_rule_id' (str): An identifier for the rule being validated.

    Raises:
        ValueError: If the file paths are a mix of directories and files, or if the 'file_regex' key
                    is used but the file paths are not directories.
    """
    for f in data["filepath"]:
        if "is_directory" in data and data["is_directory"] != f.endswith("/"):
            raise ValueError(
                "If passing a list of filepaths, all of them need to be "
                "either directories or files. Mixing is not possible. "
                "Please fix rules '{0}' filepath '{1}'".format(data["_rule_id"], f))

        data["is_directory"] = f.endswith("/")

        if "file_regex" in data and not data["is_directory"]:
            raise ValueError(
                "Used 'file_regex' key in rule '{0}' but filepath '{1}' does not "
                "specify a directory. Append '/' to the filepath or remove the "
                "'file_regex' key.".format(data["_rule_id"], f))


def enum(*args):
    """
    Creates an enumeration with the given arguments.

    Args:
        *args: A variable length argument list of strings representing the names of the
               enumeration members.

    Returns:
        type: A new enumeration type with the given member names as attributes, each assigned a
              unique integer value starting from 0.

    Example:
        >>> Colors = enum('RED', 'GREEN', 'BLUE')
        >>> Colors.RED
        0
        >>> Colors.GREEN
        1
        >>> Colors.BLUE
        2
    """
    enums = dict(zip(args, range(len(args))))
    return type('Enum', (), enums)


def recurse_or_substitute_or_do_nothing(v, string_dict, ignored_keys=frozenset()):
    """
    Recursively applies string formatting to dictionary values, substitutes strings, or returns
    the value unchanged.

    Args:
        v (Any): The value to process. Can be a dictionary, string, or any other type.
        string_dict (dict): A dictionary containing string keys and values to be used for formatting.
        ignored_keys (frozenset, optional): A set of keys to ignore when processing dictionaries.
                                            Defaults to an empty frozenset.

    Returns:
        Any: The processed value. If `v` is a dictionary, returns a dictionary with formatted values.
             If `v` is a string, returns the formatted string. Otherwise, returns `v` unchanged.
    """
    if isinstance(v, dict):
        return apply_formatting_on_dict_values(v, string_dict, ignored_keys)
    elif isinstance(v, str):
        return v.format(**string_dict)
    else:
        return v


def apply_formatting_on_dict_values(source_dict, string_dict, ignored_keys=frozenset()):
    """
    Apply string formatting on dictionary values.

    This function uses Python's built-in string replacement to replace strings marked by {token}
    if "token" is a key in the string_dict parameter. It skips keys in source_dict which are
    listed in the ignored_keys parameter. This function works only for dictionaries whose values
    are dictionaries or strings.

    Args:
        source_dict (dict): The source dictionary whose values need formatting.
        string_dict (dict): A dictionary containing replacement strings for tokens.
        ignored_keys (frozenset, optional): A set of keys to be ignored during formatting.
                                            Defaults to an empty frozenset.

    Returns:
    dict: A new dictionary with formatted values.
   """
    new_dict = {}
    for k, v in source_dict.items():
        if k not in ignored_keys:
            new_dict[k] = recurse_or_substitute_or_do_nothing(
                v, string_dict, ignored_keys)
        else:
            new_dict[k] = v
    return new_dict


def ensure_file_paths_and_file_regexes_are_correctly_defined(data):
    """
    Ensures that the data structure for file paths and file regexes is correctly defined.

    This function is used for the file_owner, file_groupowner, and file_permissions templates.
    It ensures that:
    - The 'filepath' item in the data is a list.
    - The number of items in 'file_regex' matches the number of items in 'filepath'.

    Note:
        - If 'filepath' is a string, it will be converted to a list containing that string.
        - If 'file_regex' is a string, it will be converted to a list with the same length as
          'filepath', with each element being the original 'file_regex' string.
        - If there are multiple regexes for a single filepath, the filepath must be declared
          multiple times.

    Args:
        data (dict): A dictionary containing file path and file regex information.
                     It must contain the keys:
                     - 'filepath': A string or list of strings representing file paths.
                     - 'file_regex' (optional): A string or list of strings representing file regexes.
                     - '_rule_id': A string representing the rule ID.

    Raises:
        ValueError: If the number of items in 'file_regex' does not match the number of items in 'filepath'.
    """
    # this avoids code duplicates
    if isinstance(data["filepath"], str):
        data["filepath"] = [data["filepath"]]

    if "file_regex" in data:
        # we can have a list of filepaths, but only one regex
        # instead of declaring the same regex multiple times
        if isinstance(data["file_regex"], str):
            data["file_regex"] = [data["file_regex"]] * len(data["filepath"])

        # if the length of filepaths and file_regex are not the same, then error.
        # in case we have multiple regexes for just one filepath, than we need
        # to declare that filepath multiple times
        if len(data["filepath"]) != len(data["file_regex"]):
            raise ValueError(
                "You should have one file_path per file_regex. Please check "
                "rule '{0}'".format(data["_rule_id"]))

    check_conflict_regex_directory(data)



"""
This module contains common code shared by utils/rule_dir_stats.py and utils/rule_dir_diff.py.
This code includes functions for walking the output of the utils/rule_dir_json.py script, and
filtering functions used in both scripts.
"""

from __future__ import absolute_import
from __future__ import print_function

import os
from collections import defaultdict

from .build_remediations import REMEDIATION_TO_EXT_MAP as REMEDIATION_MAP
from .utils import subset_dict


def get_affected_products(rule_obj):
    """
    Extracts and returns the set of affected products from a given rule object.

    Args:
        rule_obj (dict): A dictionary representing a rule, which contains a 'products' key.

    Returns:
        set: A set of products affected by the rule.
    """
    return set(rule_obj['products'])


def get_all_affected_products(args, rule_obj):
    """
    From a rule_obj, return the set of affected products from rule.yml, and all fixes and checks.

    If args.strict is set, this function is equivalent to get_affected_products. Otherwise, it
    includes ovals and fix content based on the values of args.fixes_only and args.ovals_only.

    Args:
        args (Namespace): The arguments passed to the script, containing flags such as strict,
                          fixes_only, and ovals_only.
        rule_obj (dict): The rule object containing information about the rule, including affected
                         products, oval products, and remediation products.

    Returns:
        set: A set of affected products based on the rule object and the provided arguments.
    """
    affected_products = get_affected_products(rule_obj)

    if args.strict:
        return affected_products

    if not args.fixes_only:
        for product in rule_obj['oval_products']:
            affected_products.add(product)

    if not args.ovals_only:
        for product in rule_obj['remediation_products']:
            affected_products.add(product)

    return affected_products


def _walk_rule(args, rule_obj, oval_func, remediation_func, verbose_output):
    """
    Walks a single rule and updates verbose_output if visited.

    Internal function for walk_rules and walk_rules_parallel.

    Args:
        args: An object containing arguments and configurations.
        rule_obj: A dictionary representing the rule to be processed.
        oval_func: A function to process the rule's OVAL definitions.
        remediation_func: A function to process the rule's remediation scripts.
        verbose_output: A dictionary to store detailed output for each rule.

    Returns:
        bool: True if the rule was visited and processed, False otherwise.
    """
    rule_id = rule_obj['id']

    affected_products = get_all_affected_products(args, rule_obj)
    if not affected_products.intersection(args.products):
        return False
    if args.query and rule_id not in args.query:
        return False

    if not args.fixes_only:
        result = oval_func(rule_obj)
        if result:
            verbose_output[rule_id]['oval'] = result

    if not args.ovals_only:
        for r_type in REMEDIATION_MAP:
            result = remediation_func(rule_obj, r_type)
            if result:
                verbose_output[rule_id][r_type] = result

    return True


def walk_rules(args, known_rules, oval_func, remediation_func):
    """
    Walks through a dictionary of known rules, conditionally calling provided functions to
    generate OVAL and remediation content, and returns the number of visited rules along with the
    output for each visited rule.

    Args:
        args (object): An object containing arguments that control the behavior of the function.
        known_rules (dict): A dictionary where keys are rule IDs and values are rule objects.
        oval_func (function): A function to be called for each rule to generate OVAL content.
        remediation_func (function): A function to be called for each rule to generate remediation content.

    Returns:
        tuple:
            affected_rules (int): The number of rules that were visited.
            verbose_output (dict): A dictionary containing the output for each visited rule.

    The output structure is a dict as follows::

        {
            rule_id: {
                "oval": oval_func(args, rule_obj),
                "ansible": remediation_func(args, "ansible", rule_obj),
                "anaconda": remediation_func(args, "anaconda", rule_obj),
                "bash": remediation_func(args, "bash", rule_obj),
                "puppet": remediation_func(args, "puppet", rule_obj)
            },
            ...
        }

    The arguments supplied to oval_func are args and rule_obj.
    The arguments supplied to remediation_func are args, the remediation type, and rule_obj.
    The input rule_obj structure is the value of known_rules[rule_id].
    """
    affected_rules = 0
    verbose_output = defaultdict(lambda: defaultdict(lambda: None))

    for rule_id in known_rules:
        rule_obj = known_rules[rule_id]
        if _walk_rule(args, rule_obj, oval_func, remediation_func, verbose_output):
            affected_rules += 1

    return affected_rules, verbose_output


def walk_rule_stats(rule_output):
    """
    Walk the output of a rule, generating statistics about affected ovals, remediations, and
    generating verbose output in a stable order.

    Args:
        rule_output (dict): The output of a rule containing information about ovals and remediations.

    Returns:
        tuple: A tuple containing the following elements:
            - affected_ovals (int): The number of affected ovals.
            - affected_remediations (int): The number of affected remediations.
            - all_affected_remediations (int): The number of rules where all remediations are affected.
            - affected_remediations_type (defaultdict): A dictionary with the count of each type of affected remediation.
            - all_output (list): A list of all affected ovals and remediations in a stable order.
    """
    affected_ovals = 0
    affected_remediations = 0
    all_affected_remediations = 0
    affected_remediations_type = defaultdict(lambda: 0)
    all_output = []

    affected_remediation = False
    all_remedation = True

    if 'oval' in rule_output:
        affected_ovals += 1
        all_output.append(rule_output['oval'])

    for r_type in sorted(REMEDIATION_MAP):
        if r_type in rule_output:
            affected_remediation = True
            affected_remediations_type[r_type] += 1
            all_output.append(rule_output[r_type])
        else:
            all_remedation = False

    if affected_remediation:
        affected_remediations += 1
    if all_remedation:
        all_affected_remediations += 1

    return (affected_ovals, affected_remediations, all_affected_remediations,
            affected_remediations_type, all_output)


def walk_rules_stats(args, known_rules, oval_func, remediation_func):
    """
    Walk a dictionary of known_rules and generate simple aggregate statistics for all visited rules.

    The oval_func and remediation_func arguments behave according to walk_rules().
    An effort is made to provide consistently ordered verbose_output by sorting all visited keys
    and the keys of ssg.build_remediations.REMEDIATION_MAP.

    Args:
        args: Arguments passed to the function.
        known_rules (dict): A dictionary of known rules to be processed.
        oval_func (function): Function to process OVAL definitions.
        remediation_func (function): Function to process remediations.

    Returns:
        tuple: A tuple containing:
            - affected_rules (int): Number of affected rules.
            - affected_ovals (int): Number of affected OVAL definitions.
            - affected_remediations (int): Number of affected remediations.
            - all_affected_remediations (int): Total number of affected remediations.
            - affected_remediations_type (dict): Dictionary with the count of each type of affected remediation.
            - all_output (list): Ordered output of all functions.
    """
    affected_rules, verbose_output = walk_rules(args, known_rules, oval_func, remediation_func)

    affected_ovals = 0
    affected_remediations = 0
    all_affected_remediations = 0
    affected_remediations_type = defaultdict(lambda: 0)
    all_output = []

    for rule_id in sorted(verbose_output):
        rule_output = verbose_output[rule_id]
        results = walk_rule_stats(rule_output)

        affected_ovals += results[0]
        affected_remediations += results[1]
        all_affected_remediations += results[2]
        for key in results[3]:
            affected_remediations_type[key] += results[3][key]

        all_output.extend(results[4])

    return (affected_rules, affected_ovals, affected_remediations,
            all_affected_remediations, affected_remediations_type, all_output)


def walk_rules_parallel(args, left_rules, right_rules, oval_func, remediation_func):
    """
    Walks two sets of known_rules (left_rules and right_rules) with identical keys and returns
    left_only, right_only, and common_only output from _walk_rule.

    If the outputted data for a rule when called on left_rules and right_rules is the same, it is
    added to common_only. Only rules which output different data will have their data added to
    left_only and right_only respectively.

    Args:
        args: Arguments to be passed to the _walk_rule function.
        left_rules (dict): Dictionary of rules on the left side.
        right_rules (dict): Dictionary of rules on the right side.
        oval_func (function): Function to process OVAL definitions.
        remediation_func (function): Function to process remediation scripts.

    Returns:
        tuple: A tuple containing three elements:
            - left_only (tuple): A tuple containing the count of affected rules and the verbose
              output for rules only in left_rules.
            - right_only (tuple): A tuple containing the count of affected rules and the verbose
              output for rules only in right_rules.
            - common_only (tuple): A tuple containing the count of affected rules and the verbose
              output for rules common to both left_rules and right_rules.

    Raises:
        AssertionError: If the sets of keys in left_rules and right_rules are not identical.
    """
    left_affected_rules = 0
    right_affected_rules = 0
    common_affected_rules = 0

    left_verbose_output = defaultdict(lambda: defaultdict(lambda: None))
    right_verbose_output = defaultdict(lambda: defaultdict(lambda: None))
    common_verbose_output = defaultdict(lambda: defaultdict(lambda: None))

    assert set(left_rules) == set(right_rules)

    for rule_id in left_rules:
        left_rule_obj = left_rules[rule_id]
        right_rule_obj = right_rules[rule_id]

        if left_rule_obj == right_rule_obj:
            if _walk_rule(args, left_rule_obj, oval_func, remediation_func, common_verbose_output):
                common_affected_rules += 1
        else:
            left_temp = defaultdict(lambda: defaultdict(lambda: None))
            right_temp = defaultdict(lambda: defaultdict(lambda: None))

            left_ret = _walk_rule(args, left_rule_obj, oval_func, remediation_func, left_temp)
            right_ret = _walk_rule(args, right_rule_obj, oval_func, remediation_func, right_temp)

            if left_ret == right_ret and left_temp == right_temp:
                common_verbose_output.update(left_temp)
                if left_ret:
                    common_affected_rules += 1
            else:
                left_verbose_output.update(left_temp)
                right_verbose_output.update(right_temp)
                if left_ret:
                    left_affected_rules += 1
                if right_ret:
                    right_affected_rules += 1

    left_only = (left_affected_rules, left_verbose_output)
    right_only = (right_affected_rules, right_verbose_output)
    common_only = (common_affected_rules, common_verbose_output)

    return left_only, right_only, common_only


def walk_rules_diff(args, left_rules, right_rules, oval_func, remediation_func):
    """
    Walk through two dictionaries of known rules and generate five sets of output.

    Does not understand renaming of rule_ids as this would depend on disk content to reflect these
    differences. Unless significantly more data is added to the rule_obj structure (contents of
    rule.yml, ovals, remediations, etc.), all information besides 'title' is not uniquely
    identifying or could be easily updated.

    Args:
        args: Arguments to be passed to the walk_rules and walk_rules_parallel functions.
        left_rules (dict): Dictionary of rules on the left side.
        right_rules (dict): Dictionary of rules on the right side.
        oval_func (function): Function to process OVAL definitions.
        remediation_func (function): Function to process remediation scripts.

    Returns:
        tuple: A five-tuple containing:
            - left_only_data: Data for rules only in left_rules.
            - right_only_data: Data for rules only in right_rules.
            - left_changed_data: Data for rules in both left_rules and right_rules but changed in left_rules.
            - right_changed_data: Data for rules in both left_rules and right_rules but changed in right_rules.
            - common_data: Data for rules common to both left_rules and right_rules.
    """
    left_rule_ids = set(left_rules)
    right_rule_ids = set(right_rules)

    left_only_rule_ids = left_rule_ids.difference(right_rule_ids)
    right_only_rule_ids = right_rule_ids.difference(left_rule_ids)
    common_rule_ids = left_rule_ids.intersection(right_rule_ids)

    left_restricted = subset_dict(left_rules, left_only_rule_ids)
    left_common = subset_dict(left_rules, common_rule_ids)
    right_restricted = subset_dict(right_rules, right_only_rule_ids)
    right_common = subset_dict(right_rules, common_rule_ids)

    left_only_data = walk_rules(args, left_restricted, oval_func, remediation_func)
    right_only_data = walk_rules(args, right_restricted, oval_func, remediation_func)
    l_c_d, r_c_d, c_d = walk_rules_parallel(args, left_common, right_common,
                                            oval_func, remediation_func)

    left_changed_data = l_c_d
    right_changed_data = r_c_d
    common_data = c_d

    return (left_only_data, right_only_data, left_changed_data, right_changed_data, common_data)


def walk_rules_diff_stats(results):
    """
    Takes the results of walk_rules_diff (results) and generates five sets of output statistics.

    Args:
        results (list): A list of five elements, where each element is a tuple containing affected
                        rules and verbose output.

    Returns:
        tuple: A tuple containing five elements, each representing the statistics for left_only
               rules, right_only rules, shared left rules, shared right rules, and shared common
               rules. Each element in the tuple is itself a tuple containing:
               - affected_rules (int): Number of affected rules.
               - affected_ovals (int): Number of affected OVAL definitions.
               - affected_remediations (int): Number of affected remediations.
               - all_affected_remediations (int): Total number of affected remediations.
               - affected_remediations_type (dict): Dictionary with remediation types as keys and
                                                    counts as values.
               - all_output (list): List of all output data.

    Raises:
        AssertionError: If the length of results is not 5 or the length of output_data is not 5.
    """
    assert len(results) == 5

    output_data = []

    for data in results:
        affected_rules, verbose_output = data

        affected_ovals = 0
        affected_remediations = 0
        all_affected_remediations = 0
        affected_remediations_type = defaultdict(lambda: 0)
        all_output = []

        for rule_id in sorted(verbose_output):
            rule_output = verbose_output[rule_id]
            _results = walk_rule_stats(rule_output)

            affected_ovals += _results[0]
            affected_remediations += _results[1]
            all_affected_remediations += _results[2]
            for key in _results[3]:
                affected_remediations_type[key] += _results[3][key]

            all_output.extend(_results[4])

        output_data.append((affected_rules, affected_ovals,
                            affected_remediations, all_affected_remediations,
                            affected_remediations_type, all_output))

    assert len(output_data) == 5

    return tuple(output_data)


def filter_rule_ids(all_keys, queries):
    """
    Filters a set of keys based on a set of queries.

    A set of queries is a comma separated list of queries, where a query is either a rule id or
    a substring thereof.

    Args:
        all_keys (iterable): An iterable containing all possible keys.
        queries (str): A comma-separated list of queries, where each query is either a rule id or
                       a substring thereof. If the literal string "all" is provided, all keys are
                       returned.

    Returns:
        set: A set of keys from all_keys that match any of the queries. If queries is empty, an
             empty set is returned. If queries is "all", all keys are returned.
    """
    if not queries:
        return set()

    if queries == 'all':
        return set(all_keys)

    # We assume that all_keys is much longer than queries; this allows us to do
    # len(all_keys) iterations of size len(query_parts) instead of len(query_parts)
    # queries of size len(all_keys) -- which hopefully should be a faster data access
    # pattern due to caches but in reality shouldn't matter. Note that we have to iterate
    # over the keys in all_keys either way, because we wish to check whether query is a
    # substring of a key, not whether query is a key.
    #
    # This does have the side-effect of not having the results be ordered according to
    # their order in query_parts, so we instead, we intentionally discard order by using
    # a set. This also guarantees that our results are unique.
    results = set()
    query_parts = queries.split(',')
    for key in all_keys:
        for query in query_parts:
            if query in key:
                results.add(key)

    return results


def missing_oval(rule_obj):
    """
    For a rule object, check if it is missing an OVAL.

    Args:
        rule_obj (dict): A dictionary representing the rule object. It must contain the keys 'id'
                         and 'ovals'.

    Returns:
        str: A message indicating the rule ID that is missing all OVALs, or None if the rule has
             OVALs.
    """
    rule_id = rule_obj['id']
    check = len(rule_obj['ovals']) > 0
    if not check:
        return "\trule_id:%s is missing all OVALs" % rule_id


def missing_remediation(rule_obj, r_type):
    """
    Check if a rule object is missing a remediation of a specified type.

    Args:
        rule_obj (dict): The rule object containing rule details.
        r_type (str): The type of remediation to check for.

    Returns:
        str: A message indicating the rule ID and the missing remediation type, if the remediation
             is missing. Otherwise, returns None.
    """
    rule_id = rule_obj['id']
    check = (r_type in rule_obj['remediations'] and
             len(rule_obj['remediations'][r_type]) > 0)
    if not check:
        return "\trule_id:%s is missing %s remediations" % (rule_id, r_type)


def two_plus_oval(rule_obj):
    """
    Check if a rule object has two or more OVALs.

    Args:
        rule_obj (dict): A dictionary representing a rule object. It should have the following keys:
            - 'id' (str): The identifier of the rule.
            - 'ovals' (list): A list of OVAL identifiers associated with the rule.

    Returns:
        str: A formatted string indicating the rule ID and its associated OVALs if there are two
             or more OVALs.
    """
    rule_id = rule_obj['id']
    check = len(rule_obj['ovals']) >= 2
    if check:
        return "\trule_id:%s has two or more OVALs: %s" % (rule_id, ','.join(rule_obj['ovals']))


def two_plus_remediation(rule_obj, r_type):
    """
    Check if a rule object has two or more remediations of a specified type.

    Args:
        rule_obj (dict): The rule object containing rule details and remediations.
        r_type (str): The type of remediation to check for.

    Returns:
        str: A formatted string indicating the rule ID and the remediations if there are two or
             more of the specified type.
    """
    rule_id = rule_obj['id']
    check = (r_type in rule_obj['remediations'] and
             len(rule_obj['remediations'][r_type]) >= 2)
    if check:
        return "\trule_id:%s has two or more %s remediations: %s" % \
               (rule_id, r_type, ','.join(rule_obj['remediations'][r_type]))


def product_names_oval(rule_obj):
    """
    Checks the consistency between the product names and OVAL object names for a given rule object.

    Args:
        rule_obj (dict): A dictionary representing a rule object. It should contain:
            - 'id' (str): The identifier of the rule.
            - 'ovals' (dict): A dictionary where keys are OVAL filenames and values are
                              dictionaries containing 'products' (list of product names).

    Returns:
        str: A message indicating if there is a mismatch between the product name and OVAL object
             name. The message format is "rule_id:<rule_id> has a different product and OVALs names:
             <product> is not <oval_product>". Returns None if all product names match their
             corresponding OVAL object names.
    """
    rule_id = rule_obj['id']
    for oval_name in rule_obj['ovals']:
        if oval_name == "shared.xml":
            continue

        oval_product, _ = os.path.splitext(oval_name)
        for product in rule_obj['ovals'][oval_name]['products']:
            if product != oval_product:
                return "\trule_id:%s has a different product and OVALs names: %s is not %s" % \
                       (rule_id, product, oval_product)


def product_names_remediation(rule_obj, r_type):
    """
    Checks the consistency between the scope of platforms and the product names of the
    remediations of a given type for a rule object.

    Args:
        rule_obj (dict): A dictionary representing the rule object, which contains an 'id' key and
                         a 'remediations' key. The 'remediations' key is a dictionary where keys
                         are remediation types and values are dictionaries of remediation names
                         and their associated products.
        r_type (str): The type of remediation to check (e.g., 'bash', 'ansible').

    Returns:
        str: A message indicating the rule ID and the inconsistency found, if any.
             The message specifies the rule ID, the remediation type, the product name
             that is inconsistent, and the expected product name.
             Returns None if no inconsistencies are found.
    """
    rule_id = rule_obj['id']
    for r_name in rule_obj['remediations'][r_type]:
        r_product, _ = os.path.splitext(r_name)
        if r_product == "shared":
            continue

        for product in rule_obj['remediations'][r_type][r_name]['products']:
            if product != r_product:
                return "\trule_id:%s has a different product and %s remediation names: %s is not %s" % \
                       (rule_id, r_type, product, r_product)



#!/usr/bin/python3

import argparse
from collections import defaultdict
import os
import json
import sys
from typing import List, Dict
import yaml


import ssg.environment
import ssg.yaml

SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
BUILD_ROOT = os.path.join(SSG_ROOT, "build")
RULES_JSON = os.path.join(BUILD_ROOT, "rule_dirs.json")
BUILD_CONFIG = os.path.join(BUILD_ROOT, "build_config.yml")


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Given a reference this script will create a control file.")
    parser.add_argument("-j", "--json", type=str,
                        help=f"Path to the rule_dirs.json file. Defaults to {RULES_JSON}",
                        default=RULES_JSON)
    parser.add_argument("-p", "--product", type=str, help="Product to build the control with",
                        required=True)
    parser.add_argument("-r", "--root", type=str,
                        help=f"Path to the root of the project. Defaults to {SSG_ROOT}",
                        default=SSG_ROOT)
    parser.add_argument("-ref", "--reference", type=str,
                        help="Reference to use for the profile. Example: ospp", required=True)
    parser.add_argument("-c", "--build-config-yaml", default=BUILD_CONFIG,
                        help=f"YAML file with information about the build configuration. "
                             f"Defaults to {BUILD_CONFIG}")
    parser.add_argument("-o", "--output", type=str, required=True,
                        help=f"Path to output the control file.")
    return parser.parse_args()


def _get_rule_dirs(json_path: str) -> Dict[str, str]:
    with open(json_path, "r") as f:
        return json.load(f)


def _check_rule_dirs_path(json: str):
    if not os.path.exists(json):
        print(f"Path {json} does not exist.", file=sys.stderr)
        print("Hint: run ./utils/rule_dir_json.py first.", file=sys.stderr)
        raise SystemExit(1)


def _get_env_yaml(root: str, product: str, build_config_yaml: str) -> Dict:
    product_dir = os.path.join(root, "products", product)
    product_yaml_path = os.path.join(product_dir, "product.yml")
    env_yaml = ssg.environment.open_environment(
        build_config_yaml, product_yaml_path, os.path.join(root, "product_properties"))
    return env_yaml


def _get_id_mapping(env_yaml, reference, json_path: str) -> Dict:
    rule_dir_json: Dict = _get_rule_dirs(json_path)
    id_mapping: Dict[str, list[str]] = defaultdict(list) # type: ignore [misc] # For old mypy
    for rule_id, rule_obj in rule_dir_json.items():
        rule_yaml = os.path.join(rule_obj["dir"], "rule.yml")
        rule = ssg.yaml.open_and_macro_expand(rule_yaml, env_yaml)
        if "references" not in rule:
            continue
        ref_id = rule["references"].get(reference)
        if not ref_id:
            continue
        ids: List[str] = ref_id.split(",")
        for _id in ids:
            id_mapping[_id].append(rule_id)
    return id_mapping


def main() -> int:
    args = _parse_args()
    _check_rule_dirs_path(args.json)
    env_yaml = _get_env_yaml(args.root, args.product, args.build_config_yaml)
    id_mapping = _get_id_mapping(env_yaml, args.reference, args.json)
    output = dict()
    output["levels"] = [{'id': 'base'}]
    output["controls"] = list()
    for _id in sorted(id_mapping.keys()):
        rules = id_mapping[_id]
        control = dict()
        control["id"] = _id
        control["levels"] = ["base"]
        control["rules"] = sorted(rules)
        control["status"] = "automated"
        output["controls"].append(control)

    with open(args.output, "w") as f:
        f.write(yaml.dump(output, sort_keys=False))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/python3

import xml.etree.ElementTree as ET
import os
import os.path
import argparse

XCCDF_NS = "http://checklists.nist.gov/xccdf/1.2"


def compare_lists(rules_in_ds_with_ansible_fix, playbooks_in_dir):
    ds_set = set(rules_in_ds_with_ansible_fix)
    pb_dir_set = set(playbooks_in_dir)
    if not ds_set.issubset(pb_dir_set):
        raise Exception("Rules without playbooks: {%s}" %
                        repr(list(ds_set.difference(pb_dir_set))))


def compare_ds_with_playbooks_dir(ds_path, playbooks_dir_path):
    tree = ET.parse(ds_path)
    root = tree.getroot()
    rules_in_ds_with_ansible_fix = []
    # uses the first benchmark if multiple benchmarks are present
    benchmark = root.find(".//{%s}Benchmark" % XCCDF_NS)
    for rule in benchmark.findall(".//{%s}Rule" % XCCDF_NS):
        for fix in rule.findall("./{%s}fix" % XCCDF_NS):
            system = fix.get("system")
            if system == "urn:xccdf:fix:script:ansible":
                id_ = rule.get("id")
                id_ = id_.replace("xccdf_org.ssgproject.content_rule_", "")
                rules_in_ds_with_ansible_fix.append(id_)
    playbooks_in_dir = []
    for filename in os.listdir(playbooks_dir_path):
        id_, _ = os.path.splitext(filename)
        playbooks_in_dir.append(id_)
    compare_lists(rules_in_ds_with_ansible_fix, playbooks_in_dir)


def main():
    parser = argparse.ArgumentParser(
        description="Tests if Ansible Playbooks were generated for all rules"
        "that have an Ansible remediation available in data stream."
    )
    parser.add_argument("--build-dir", required=True,
                        help="Build directory containing built data streams"
                        "and playbooks subdirectory")
    parser.add_argument("--product", required=True,
                        help="Product ID")
    args = parser.parse_args()
    ds_path = os.path.join(args.build_dir, "ssg-" + args.product + "-ds.xml")
    playbooks_dir_path = os.path.join(args.build_dir, args.product,
                                      "playbooks", "all")
    compare_ds_with_playbooks_dir(ds_path, playbooks_dir_path)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import difflib
import re
import sys
import xml.etree.ElementTree as ET

import ssg.xml
from ssg.constants import FIX_TYPE_TO_SYSTEM, XCCDF12_NS
from ssg.utils import mkdir_p


class StandardContentDiffer(object):

    def __init__(self, old_content, new_content, rule_id, show_diffs, rule_diffs,
                 only_rules, output_dir):
        self.old_content = old_content
        self.new_content = new_content

        self.rule_id = rule_id
        self.show_diffs = show_diffs
        self.rule_diffs = rule_diffs
        self.only_rules = only_rules

        self.context_lines = 3

        self.check_system_map = {
            "OVAL": {"uri": ssg.constants.oval_namespace, "comp_func": self.compare_ovals},
            "OCIL": {"uri": ssg.constants.ocil_cs, "comp_func": self.compare_ocils}
        }

        self.output_dir = output_dir
        if self.rule_diffs:
            self._ensure_output_dir_exists()

    def _ensure_output_dir_exists(self):
        try:
            mkdir_p(self.output_dir)
        except OSError:
            print("Output path '%s' exists and it is not a directory." % self.output_dir)
            sys.exit(1)

    def output_diff(self, identifier, diff, mode="a"):
        if not diff:
            return

        if self.rule_diffs:
            with open("%s/%s" % (self.output_dir, identifier), mode) as f:
                f.write(diff)
        else:
            print(diff)

    def _get_rules_to_compare(self, benchmark):
        rule_to_find = self.rule_id
        if self.rule_id:
            if not self.rule_id.startswith(ssg.constants.OSCAP_RULE):
                rule_to_find = ssg.constants.OSCAP_RULE + self.rule_id
        rules = benchmark.find_rules(rule_to_find)
        return rules

    def compare_rules(self, old_benchmark, new_benchmark):
        try:
            rules_in_old_benchmark = self._get_rules_to_compare(old_benchmark)
        except ValueError as e:
            print(str(e))
            return

        for old_rule in rules_in_old_benchmark:
            rule_id = old_rule.get_attr("id")
            new_rule = new_benchmark.find_rule(rule_id)
            if new_rule is None:
                print("%s is missing in new data stream." % (rule_id))
                continue
            if self.only_rules:
                continue
            self.compare_rule(old_rule, new_rule, rule_id)
            self.compare_platforms(old_rule, new_rule,
                                   old_benchmark, new_benchmark, rule_id)

        if self.rule_diffs:
            print("Diff files saved at %s." % self.output_dir)

    def compare_rule(self, old_rule, new_rule, identifier):
        self.compare_rule_texts(old_rule, new_rule, identifier)
        self.compare_checks(old_rule, new_rule, "OVAL", identifier)
        self.compare_checks(old_rule, new_rule, "OCIL", identifier)
        for remediation_type in FIX_TYPE_TO_SYSTEM.keys():
            self.compare_remediations(old_rule, new_rule, remediation_type, identifier)

    def _get_list_of_platforms(self, cpe_platforms, idref):
        cpe_list = []
        if len(cpe_platforms) == 0:
            print("Platform {0} not defined in platform specification".format(idref))
            return cpe_list
        elif len(cpe_platforms) > 1:
            print("Platform {0} defined more than once".format(idref))
            for cpe_p in cpe_platforms:
                ET.dump(cpe_p)
        for cpe_platform in cpe_platforms:
            check_fact_refs = cpe_platform.find_all_check_fact_ref_elements()
            for check_fact_ref in check_fact_refs:
                cpe_list.append(check_fact_ref.get("id-ref"))
        return cpe_list

    def compare_platforms(self, old_rule, new_rule, old_benchmark, new_benchmark, identifier):
        entries = [{
                "benchmark": old_benchmark,
                "rule": old_rule,
                "cpe": []
            }, {
                "benchmark": new_benchmark,
                "rule": new_rule,
                "cpe": []
            }]

        for entry in entries:
            for platform in entry["rule"].get_all_platform_elements():
                idref = platform.get("idref")
                if idref.startswith("#"):
                    cpe_platforms = entry["benchmark"].find_all_cpe_platforms(idref)
                    entry["cpe"] += self._get_list_of_platforms(cpe_platforms, idref)
                else:
                    entry["cpe"].append(idref)

        if entries[0]["cpe"] != entries[1]["cpe"]:
            print("Platform has been changed for rule '{0}'".format(identifier))

            if self.show_diffs:
                diff = self.generate_diff_text("\n".join(entries[0]["cpe"])+"\n",
                                               "\n".join(entries[1]["cpe"])+"\n",
                                               fromfile=identifier, tofile=identifier)
                self.output_diff(identifier, diff)

    def compare_rule_texts(self, old_rule, new_rule, identifier):
        old_rule_text = old_rule.join_text_elements()
        new_rule_text = new_rule.join_text_elements()

        if old_rule_text == new_rule_text:
            return

        if old_rule_text != "":
            print(
                "New content has different text for rule '%s'." % (identifier))

        if self.show_diffs:
            diff = self.generate_diff_text(old_rule_text, new_rule_text,
                                           fromfile=identifier, tofile=identifier,
                                           n=self.context_lines)

            self.output_diff(identifier, diff, mode="w")

    def compare_check_ids(self, system, identifier, old_check_id, new_check_id):
        if old_check_id != new_check_id:
            print(
                "%s definition ID for rule '%s' has changed from "
                "'%s' to '%s'." % (
                    system, identifier, old_check_id, new_check_id)
            )

    def compare_check_file_names(self, system, identifier,
                                 old_check_file_name, new_check_file_name):
        if old_check_file_name != new_check_file_name:
            print(
                "%s definition file for rule '%s' has changed from "
                "'%s' to '%s'." % (
                    system, identifier, old_check_file_name, new_check_file_name)
            )

    def get_check_docs(self, system, identifier, old_check_file_name, new_check_file_name):
        try:
            old_check_doc = self.old_content.components.get(system)[old_check_file_name]
        except (KeyError, TypeError):
            print(
                "Rule '%s' points to '%s' which isn't a part of the "
                "old data stream" % (identifier, old_check_file_name))
            old_check_doc = None
        try:
            new_check_doc = self.new_content.components.get(system)[new_check_file_name]
        except (KeyError, TypeError):
            print(
                "Rule '%s' points to '%s' which isn't a part of the "
                "new data stream" % (identifier, new_check_file_name))
            new_check_doc = None
        return old_check_doc, new_check_doc

    def compare_checks(self, old_rule, new_rule, system, identifier):
        check_system_uri = self.check_system_map[system]["uri"]
        old_check = old_rule.get_check_element(check_system_uri)
        new_check = new_rule.get_check_element(check_system_uri)
        if (old_check is None and new_check is not None):
            print("New data stream adds %s for rule '%s'." % (system, identifier))
        elif (old_check is not None and new_check is None):
            print(
                "New data stream is missing %s for rule '%s'." % (system, identifier))
        elif (old_check is not None and new_check is not None):
            old_check_content_ref = old_rule.get_check_content_ref_element(old_check)
            new_check_content_ref = new_rule.get_check_content_ref_element(new_check)
            old_check_id = old_check_content_ref.get("name")
            new_check_id = new_check_content_ref.get("name")
            old_check_file_name = old_check_content_ref.get("href")
            new_check_file_name = new_check_content_ref.get("href")

            self.compare_check_ids(system, identifier, old_check_id, new_check_id)
            self.compare_check_file_names(system, identifier,
                                          old_check_file_name, new_check_file_name)

            if (self.show_diffs and
               identifier != "xccdf_org.ssgproject.content_rule_security_patches_up_to_date"):

                old_check_doc, new_check_doc = self.get_check_docs(system, identifier,
                                                                   old_check_file_name,
                                                                   new_check_file_name)
                if not old_check_doc or not new_check_doc:
                    return

                self.check_system_map[system]["comp_func"](old_check_doc, old_check_id,
                                                           new_check_doc, new_check_id, identifier)

    def compare_ovals(self, old_oval_def_doc, old_oval_def_id,
                      new_oval_def_doc, new_oval_def_id, identifier):
        old_def = old_oval_def_doc.find_oval_definition(old_oval_def_id)
        new_def = new_oval_def_doc.find_oval_definition(new_oval_def_id)
        old_els = old_def.get_elements()
        new_els = new_def.get_elements()

        old_els_text = self.serialize_elements(old_els)
        new_els_text = self.serialize_elements(new_els)
        diff = self.generate_diff_text(old_els_text, new_els_text,
                                       fromfile=old_oval_def_id, tofile=new_oval_def_id)
        if diff:
            print("OVAL for rule '%s' differs." % (identifier))
        self.output_diff(identifier, diff)

    def compare_ocils(self, old_ocil_doc, old_ocil_id, new_ocil_doc, new_ocil_id, identifier):
        try:
            old_question = old_ocil_doc.find_boolean_question(old_ocil_id)
            new_question = new_ocil_doc.find_boolean_question(new_ocil_id)
        except ValueError as e:
            print("Rule '%s' OCIL can't be found: %s" % (identifier, str(e)))
            return
        diff = self.generate_diff_text(old_question, new_question,
                                       fromfile=old_ocil_id, tofile=new_ocil_id)
        if diff:
            print("OCIL for rule '%s' differs." % identifier)
        self.output_diff(identifier, diff)

    def compare_remediations(self, old_rule, new_rule, remediation_type, identifier):
        system = FIX_TYPE_TO_SYSTEM[remediation_type]
        old_fix = old_rule.get_fix_element(system)
        new_fix = new_rule.get_fix_element(system)
        if (old_fix is None and new_fix is not None):
            print("New data stream adds %s remediation for rule '%s'." % (
                remediation_type, identifier))
        elif (old_fix is not None and new_fix is None):
            print("New data stream is missing %s remediation for rule '%s'." % (
                remediation_type, identifier))
        elif (old_fix is not None and new_fix is not None):
            self._compare_fix_elements(
                old_fix, new_fix, remediation_type, identifier)

    def _compare_fix_elements(self, old_fix, new_fix, remediation_type, identifier):
        old_fix_id = old_fix.get("id")
        new_fix_id = new_fix.get("id")
        if old_fix_id != new_fix_id:
            print(
                "%s remediation ID for rule '%s' has changed from "
                "'%s' to '%s'." % (
                    remediation_type, identifier, old_fix_id, new_fix_id)
            )
        if self.show_diffs:
            old_fix_text = "".join(old_fix.itertext())
            new_fix_text = "".join(new_fix.itertext())
            diff = self.generate_diff_text(old_fix_text, new_fix_text,
                                           fromfile=identifier, tofile=identifier)
            if diff:
                print("%s remediation for rule '%s' differs." % (remediation_type, identifier))
            self.output_diff(identifier, diff)

    def generate_diff_text(self, old_r, new_r,
                           fromfile="old data stream", tofile="new data stream", n=3):
        if old_r != new_r:
            diff = "".join(difflib.unified_diff(
                old_r.splitlines(keepends=True), new_r.splitlines(keepends=True),
                fromfile=fromfile, tofile=tofile, n=n))
            return diff
        return None

    def serialize_elements(self, elements):
        text = ""
        for thing, atrribute in elements:
            text += "%s %s\n" % (thing, atrribute)
        return text


class StigContentDiffer(StandardContentDiffer):

    def __init__(self, old_content, new_content, rule_id, show_diffs, rule_diffs,
                 only_rules, output_dir):
        super(StigContentDiffer, self).__init__(old_content, new_content, rule_id,
                                                show_diffs, rule_diffs, only_rules, output_dir)

        self.context_lines = 200

    def _get_stig_id(self, element):
        return element.get_version_element().text

    def get_stig_rule_SV(self, sv_rule_id):
        stig_rule_id = re.search(r'(SV-\d+)r\d+_rule', sv_rule_id)
        if not stig_rule_id:
            print("The rule '%s' doesn't have the usual STIG format: 'SV-XXXXXXrXXXXXX_rule.\n"
                  "Please make sure the input contents are DISA STIG Benchmarks." % sv_rule_id)
            sys.exit(1)
        return stig_rule_id.group(1)

    def compare_rules(self, old_benchmark, new_benchmark):
        try:
            rules_in_old_benchmark = self._get_rules_to_compare(old_benchmark)
            rules_in_new_benchmark = self._get_rules_to_compare(new_benchmark)
        except ValueError as e:
            print(str(e))
            return

        # The rules in STIG Benchmarks will have their IDS changed whenever there is an update.
        # However, only the release number changes, the SV number stays the same.
        # This creates map the SV numbers to their equivalent full IDs in the new Benchmark.
        new_rule_mapping = {self.get_stig_rule_SV(rule.get_attr("id")): rule.get_attr("id")
                            for rule in rules_in_new_benchmark}
        old_rule_mapping = {self.get_stig_rule_SV(rule.get_attr("id")): rule.get_attr("id")
                            for rule in rules_in_old_benchmark}

        self.compare_existing_rules(new_benchmark,
                                    old_benchmark, rules_in_old_benchmark, new_rule_mapping)

        self.check_for_new_rules(rules_in_new_benchmark, old_rule_mapping)

        if self.rule_diffs:
            print("Diff files saved at %s." % self.output_dir)

    def compare_existing_rules(self, new_benchmark, old_benchmark,
                               rules_in_old_benchmark, new_rule_mapping):
        for old_rule in rules_in_old_benchmark:
            old_sv_rule_id = self.get_stig_rule_SV(old_rule.get_attr("id"))
            old_stig_id = self._get_stig_id(old_rule)
            try:
                new_sv_rule_id = new_rule_mapping[old_sv_rule_id]
            except KeyError:
                print("%s is missing in new data stream." % old_stig_id)
                continue
            if self.only_rules:
                continue

            new_rule = new_benchmark.find_rule(new_sv_rule_id)
            new_stig_id = self._get_stig_id(new_rule)

            self.compare_rule(old_rule, new_rule, new_stig_id)
            self.compare_platforms(old_rule, new_rule,
                                   old_benchmark, new_benchmark, new_stig_id)

    def check_for_new_rules(self, rules_in_new_benchmark, old_rule_mapping):
        # Check for rules added in new content
        for new_rule in rules_in_new_benchmark:
            new_stig_id = self._get_stig_id(new_rule)
            new_sv_rule_id = self.get_stig_rule_SV(new_rule.get_attr("id"))
            try:
                old_sv_rule_id = old_rule_mapping[new_sv_rule_id]
            except KeyError:
                print("%s was added in new data stream." % (new_stig_id))

                # Compare against empty rule so that a diff is generated
                empty_rule = ssg.xml.XMLRule(ET.Element("{%s}Rule" % XCCDF12_NS))
                self.compare_rule(empty_rule, new_rule, new_stig_id)




#!/usr/bin/env python3

import openpyxl
from openpyxl.styles import Alignment, Font, PatternFill
from openpyxl.styles.colors import Color

import utils.srg_export.data

MICRO_COLUMN_SIZE = 8
SMALL_COLUMN_SIZE = 17
MEDIUM_COLUMN_SIZE = SMALL_COLUMN_SIZE*1.5
BIG_COLUMN_SIZE = SMALL_COLUMN_SIZE*3
HUGE_COLUMN_SIZE = SMALL_COLUMN_SIZE*4
COLUMN_SIZES = {
    'IA Control': SMALL_COLUMN_SIZE,
    'CCI': SMALL_COLUMN_SIZE,
    'SRGID': SMALL_COLUMN_SIZE,
    'STIGID': MICRO_COLUMN_SIZE,
    'SRG Requirement': HUGE_COLUMN_SIZE,
    'Requirement': HUGE_COLUMN_SIZE,
    'SRG VulDiscussion': HUGE_COLUMN_SIZE,
    'Vul Discussion': HUGE_COLUMN_SIZE,
    'Status': MEDIUM_COLUMN_SIZE,
    'SRG Check': HUGE_COLUMN_SIZE,
    'Check': HUGE_COLUMN_SIZE,
    'SRG Fix': MICRO_COLUMN_SIZE,
    'Fix': HUGE_COLUMN_SIZE,
    'Severity': MICRO_COLUMN_SIZE,
    'Mitigation': BIG_COLUMN_SIZE,
    'Artifact Description': BIG_COLUMN_SIZE,
    'Status Justification': BIG_COLUMN_SIZE
}


def setup_sheet(sheet: openpyxl.worksheet.worksheet.Worksheet) -> None:
    for column, header in utils.srg_export.data.COLUMN_MAPPINGS.items():
        sheet.column_dimensions[f'{column}'].width = COLUMN_SIZES[header]


def setup_headers(sheet: openpyxl.worksheet.worksheet.Worksheet) -> None:
    for column, header in utils.srg_export.data.COLUMN_MAPPINGS.items():
        sheet[f'{column}1'] = header
    for cell in list(sheet.iter_rows(max_row=1))[0]:
        cell.font = Font(bold=True, name='Calibri')


def format_cells(sheet: openpyxl.worksheet.worksheet.Worksheet):
    for row in sheet.iter_rows():
        for cell in row:
            cell.alignment = Alignment(wrap_text=True, vertical='top')


def setup_row(sheet: openpyxl.worksheet.worksheet.Worksheet, row: dict, row_num: int) -> None:
    for column, header in utils.srg_export.data.COLUMN_MAPPINGS.items():
        sheet[f'{column}{row_num}'] = row[header]
    sheet.row_dimensions[row_num].height = 130
    if row_num > 1 and ('Fix' not in row or not row['Fix']):
        highlight_row(sheet, row_num, 23)

    # freeze header row represented by A1
    # A2 is required because it freezes everything before it
    # in this case we only want A1 row to be frozen
    sheet.freeze_panes = "A2"


def highlight_row(sheet: openpyxl.worksheet.worksheet.Worksheet, row_num: int, color: int) -> None:
    row = list(sheet.iter_rows(min_row=row_num, max_row=row_num, min_col=1))[0]
    for cell in row:
        cell.fill = PatternFill(start_color=Color(indexed=color), end_color=Color(indexed=color),
                                fill_type="solid")


def handle_dict(data: list, output_path: str, sheet_name: str) -> None:
    """
    Given a dict with the fields for the srg export, create a formatted XLSX file
    """
    xlsx = openpyxl.Workbook()
    sheet = xlsx.active
    sheet.name = sheet_name
    setup_headers(sheet)
    setup_sheet(sheet)
    row_num = 2
    for row in data:
        setup_row(sheet, row, row_num)
        row_num += 1
    format_cells(sheet)
    xlsx.save(output_path)
